#
# This file is part of the GROMACS molecular simulation package.
#
# Copyright 2019- The GROMACS Authors
# and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
# Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
#
# GROMACS is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# GROMACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with GROMACS; if not, see
# https://www.gnu.org/licenses, or write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
#
# If you want to redistribute modifications to GROMACS, please
# consider that scientific software is very special. Version
# control is crucial - bugs must be traceable. We will be happy to
# consider code for inclusion in the official distribution, but
# derived work must not be called official GROMACS. Details are found
# in the README & COPYING files - if they are missing, get the
# official version at https://www.gromacs.org.
#
# To help us fund GROMACS development, we humbly ask that you cite
# the research papers on the package. Check out https://www.gromacs.org.

# This file is based on the Kasson Lab gmxapi project release 0.0.7.4.
# https://github.com/kassonlab/gmxapi/blob/v0.0.7.4/src/gmx/context.py
# https://github.com/kassonlab/gmxapi/blob/v0.0.7.4/LICENSE

"""
Execution Context
=================
"""

__all__ = ["Context"]

import importlib
import os
import warnings
import tempfile
from typing import Optional, TYPE_CHECKING, Any

import numpy as np

import gmxapi._gmxapi as _gmxapi
import gmxapi.utility
from gmxapi import exceptions
from gmxapi import logger as root_logger
from gmxapi.exceptions import ApiError
from gmxapi.exceptions import DataShapeError
from gmxapi.exceptions import ProtocolError

if TYPE_CHECKING:
    from gmxapi.runtime import (
        ResourceAllocation,
        ResourceAssignment,
    )
    import mpi4py.MPI

# Module-level logger
logger = root_logger.getChild("simulation.context")
logger.info("Importing {}".format(__name__))


def _load_tpr(self, element):
    """Implement the gromacs.load_tpr operation.

    Updates the minimum width of the workflow parallelism. Does not add any API object to the graph.

    Arguments:
        self: The Context in which this operation is being loaded.
        element: WorkElement specifying the operation.

    Returns:
        A Director that the Context can use in launching the Session.
    """

    class Builder(object):
        def __init__(self, tpr_list):
            logger.debug("Loading tpr builder for tpr_list {}".format(tpr_list))
            self.tpr_list = tpr_list
            self.subscribers = []
            self.width = len(tpr_list)

        def add_subscriber(self, builder):
            builder.infile = self.tpr_list
            self.subscribers.append(builder)

        def build(self, dag):
            width = len(self.tpr_list)
            for builder in self.subscribers:
                builder.width = width
            if "width" in dag.graph:
                width = max(width, dag.graph["width"])
            dag.graph["width"] = width

    return Builder(element.params["input"])


def _md(context, element):
    """Implement the gmxapi.md operation by returning a builder that can populate a data flow graph for the element.

    Inspects dependencies to set up the simulation runner.

    The graph node created will have `launch` and `run` attributes with function references, and a `width`
    attribute declaring the workflow parallelism requirement.

    Arguments:
        context: The Context in which this operation is being loaded.
        element: WorkElement specifying the operation.

    Returns:
        A Director that the Context can use in launching the Session.
    """

    class Builder(object):
        """Translate md work element to a node in the session's DAG."""

        def __init__(self, element):
            try:
                self.name = element.name
                # Note that currently the calling code is in charge of subscribing this builder to its dependencies.
                # A list of tpr files will be set when the calling code subscribes this builder to a tpr provider.
                self.infile = None
                # Other dependencies in the element may register potentials when subscribed to.
                self.potential = []
                self.input_nodes = []
                self.runtime_params = element.params
            except AttributeError:
                raise exceptions.ValueError(
                    "object provided does not seem to be a WorkElement."
                )

        def add_subscriber(self, builder):
            """The md operation does not yet have any subscribeable facilities."""
            pass

        def build(self, dag):
            """Add a node to the graph that, when launched, will construct a simulation runner.

            Complete the definition of appropriate graph edges for dependencies.

            The launch() method of the added node creates the runner from the tpr file for the current rank and adds
            modules from the incoming edges.
            """
            if not (
                hasattr(dag, "add_node")
                and hasattr(dag, "add_edge")
                and hasattr(dag, "graph")
                and hasattr(dag, "nodes")
            ):
                raise exceptions.ProtocolError(
                    "Argument 'dag' must provide the DiGraph interface."
                )
            name = self.name
            dag.add_node(name)
            for neighbor in self.input_nodes:
                dag.add_edge(neighbor, name)
            infile = self.infile
            assert infile is not None
            potential_list = self.potential

            # We may have to widen the graph according to runtime_params.
            # We also ensure that the ensemble data is represented as a list of dict.
            if isinstance(self.runtime_params, list):
                runtime_params = self.runtime_params
                params_width = len(runtime_params)
            else:
                assert isinstance(self.runtime_params, dict)
                # First, determine the final width
                params_width = 1
                for key, value in self.runtime_params.items():
                    if isinstance(value, list):
                        value_length = len(value)
                        if params_width == 1:
                            params_width = value_length
                        elif (value_length > 1) and (params_width != value_length):
                            raise DataShapeError(
                                f"runtime parameter {key}={value} not compatible with ensemble width "
                                f"{params_width}"
                            )
                # Get the list of flattened params dicts
                runtime_params = [dict() for _ in range(params_width)]
                for param, value in self.runtime_params.items():
                    if not isinstance(value, list):
                        for member in range(params_width):
                            runtime_params[member][param] = value
                    elif len(value) == 1:
                        value = value[0]
                        for member in range(params_width):
                            runtime_params[member][param] = value
                    else:
                        assert isinstance(value, list)
                        value_length = len(value)
                        assert value_length > 1
                        assert value_length == params_width
                        for member in range(params_width):
                            runtime_params[member][param] = value[member]
            assert isinstance(runtime_params, list)
            assert params_width == len(runtime_params)

            input_width = len(infile)

            dag_width = dag.graph["width"]
            required_width = max((dag_width, input_width, params_width))

            if required_width > 1:
                if (dag_width != 1) and (dag_width != required_width):
                    raise DataShapeError(
                        f"Inputs {infile} and params {runtime_params} cannot be fit to current data flow "
                        f"graph of width {dag_width}."
                    )
                if input_width == 1:
                    infile = [infile[0]] * required_width
                elif input_width != required_width:
                    raise DataShapeError(
                        f"Inputs {infile} incompatible with required ensemble width {required_width}."
                    )
                if params_width == 1:
                    _params: dict = runtime_params[0]
                    runtime_params = [_params.copy() for _ in range(required_width)]
                elif params_width != required_width:
                    raise DataShapeError(
                        f"Params {runtime_params} incompatible with required ensemble width {required_width}."
                    )

            self.infile = infile
            self.runtime_params = runtime_params
            dag.graph["width"] = required_width

            # Provide closure with which to execute tasks for this node.
            # TODO(#4079): launch is explicitly a collaboration with the Context
            #  to build a Session, and should be decoupled from element.workspec._context
            def launch(ensemble_member=None):
                assert ensemble_member is not None
                # Note that gmxapi 0.0.7 does not support any consumers of mdrun, so the graph cannot be
                # further widened, and launch can assume that *rank* <= dag width.
                assert dag.graph["width"] == required_width
                assert len(self.infile) == required_width
                assert len(self.runtime_params) == required_width
                if ensemble_member > required_width:
                    raise ApiError(
                        f"Trying to launch MD for simulator {ensemble_member}, "
                        f"but ensemble has width {required_width}."
                    )
                infile = self.infile[ensemble_member]
                runtime_params = self.runtime_params[ensemble_member]

                # Copy and update, if required by `end_time` parameter.
                temp_filename = None
                if "end_time" in runtime_params:
                    # Note that mkstemp returns a file descriptor as the first part of the tuple.
                    # We can make this cleaner in 0.0.7 with a separate node that manages the
                    # altered input.
                    _, temp_filename = tempfile.mkstemp(suffix=".tpr")
                    logger.debug(
                        "Updating input. Using temp file {}".format(temp_filename)
                    )
                    _gmxapi.rewrite_tprfile(
                        source=infile,
                        destination=temp_filename,
                        end_time=runtime_params["end_time"],
                    )
                    tpr_file = temp_filename
                else:
                    tpr_file = infile

                logger.info("Loading TPR file: {}".format(tpr_file))
                system = _gmxapi.from_tpr(tpr_file)
                dag.nodes[name]["system"] = system
                mdargs = _gmxapi.MDArgs()
                mdargs.set(runtime_params)
                # Workaround to give access to plugin potentials used in a context.
                # See https://gitlab.com/gromacs/gromacs/-/issues/3145
                work_context: Context = element.workspec._context
                work_context.potentials = potential_list
                pycontext: _gmxapi.Context = work_context._api_context
                pycontext.setMDArgs(mdargs)
                for potential in potential_list:
                    pycontext.add_mdmodule(potential)
                dag.nodes[name]["session"] = system.launch(pycontext)
                dag.nodes[name]["close"] = dag.nodes[name]["session"].close

                if "end_time" in runtime_params:

                    def special_close():
                        dag.nodes[name]["session"].close()
                        logger.debug(
                            "Unlinking temporary TPR file {}.".format(temp_filename)
                        )
                        os.unlink(temp_filename)

                    dag.nodes[name]["close"] = special_close
                else:
                    dag.nodes[name]["close"] = dag.nodes[name]["session"].close

                def runner():
                    """Currently we only support a single call to run."""

                    def done():
                        raise StopIteration()

                    # Replace the runner with a stop condition for subsequent passes.
                    dag.nodes[name]["run"] = done
                    return dag.nodes[name]["session"].run()

                dag.nodes[name]["run"] = runner
                return dag.nodes[name]["run"]

            dag.nodes[name]["launch"] = launch

    return Builder(element)


class Context(object):
    """Manage an array of simulation work executing in parallel.

    .. deprecated:: 0.0.7

    This execution context implementation is imported from the gmxapi 0.0.7
    package for GROMACS 2019 and does not conform to the protocols of gmxapi 0.1+.
    It is used internally to support legacy code.

    The following features are subject to be changed or removed without further
    notice.

    Attributes:
        work :obj:`gmx.workflow.WorkSpec`: specification of work to be performed when a session is launched.
        ensemble_rank : numerical index of the simulator task in an ensemble (for participating ranks).
        simulator_rank : numerical index of the current worker in a running simulator (None if not running).
        work_width : Detected number of co-scheduled work elements.
        elements : dictionary of references to elements of the workflow.

    *rank*, *work_width*, and *elements* are empty or None until the work is
    processed, as during session launch.

    To run multiple simulations at a time, whether ensembles or independent
    simulations, Python should be invoked in an MPI environment with `mpi4py`.

    Example:

        >>> from gmxapi.simulation.context import get_context
        >>> from gmxapi.simulation.workflow import from_tpr
        >>> work = from_tpr([tpr_filename, tpr_filename])
        >>> context = get_context(work)
        >>> with context as session:
        ...    session.run()
        ...    # The session is one abstraction too low to know what rank it is. It lets the spawning context manage
        ...    # such things.
        ...    # rank = session.rank
        ...    # The local context object knows where it fits in the global array.
        ...    member = context.ensemble_rank
        ...    output_path = os.path.join(context.workdir_list[rank], 'traj.trr')
        ...    assert(os.path.exists(output_path))
        ...    print('Worker {} produced {}'.format(rank, output_path))

    Warning:
        Do not run a gmxapi script in an MPI environment without `mpi4py`. gmxapi
        will not be able to determine that other processes are running the same
        script.

    Implementation notes:

    To produce a running session, the Context __enter__() method is called,
    according to the Python context manager protocol. At this time, the attached
    WorkSpec must be feasible  on the available resources. To turn the specified
    work into an executable directed acyclic graph (DAG), handle objects for the
    elements in the work spec are sequenced in dependency-compatible order and
    the context creates a "builder" for each according to the element's operation.
    Each builder is subscribed to the builders of its dependency elements. The
    DAG is then assembled by calling each builder in sequence. A builder can add
    zero, one, or more nodes and edges to the DAG.

    The Session is then launched from the DAG. What happens next is
    implementation-dependent, and it may take a while for us to decide whether
    and how to standardize interfaces for the DAG nodes and edges and/or execution
    protocols. I expect each node will at least have a `launch()` method, but will
    also probably have typed input and output ports as well as some signalling.
    A sophisticated and abstract Session implementation could schedule work only
    to satisfy data dependencies of requested output upon request. Our immediate
    implementation will use the following protocol.

    Each node has a `launch()` method. When the session is entered, the `launch()`
    method is called for each node in dependency order. The launch method returns
    either a callable (`run()` function) or None, raising an exception in
    case of an error. The sequence of callables is stored by the Session. When
    Session.run() is called, the sequence of callables is called in order. If
    StopIteration is raised by the callable, it is removed from the sequence.
    The sequence is processed repeatedly until there are no more callables.

    Note that this does not rigorously handle races or deadlocks, or flexibility
    in automatically chasing dependencies. A more thorough implementation could
    recursively call launch on dependencies (launch could be idempotent or involve
    some signalling to dependents when complete), run calls could be entirely
    event driven, and/or nodes could "publish" output (including just a completion
    message), blocking for acknowledgement before looking for the next set of
    subscribed inputs.
    """

    _api_context: Optional[_gmxapi.Context] = None
    """gmxapi library Context object for managing GROMACS resources.
    
    .. versionchanged:: 0.4
    
        Only processes (MPI ranks) participating in the simulation work will
        acquire an API Context. MPI-enabled GROMACS checks the validity of the
        communicator provided to the Context (and will not issue an unusable
        Context object). It is easier to handle a ``None`` object than to
        develop and use some sort of mock or dummy context.

    """

    __workdir_list: Optional[list] = None
    """Caller-provided working directories (if any).
    
    Support the (deprecated) *workdir_list* key word argument.
    """

    __session_allocation: "Optional[ResourceAllocation[mpi4py.MPI.Comm, Any]]" = None
    """Base resources (if any) from which to assign task resources.
    
    To be removed when we no longer accept *communicator* instead of *resources*.
    """

    _session_resources: "Optional[ResourceAssignment[mpi4py.MPI.Comm, Any]]" = None
    """Session scoped communication resources (where applicable).
    
    When the Context is used as a Python context manager, the state of the
    Context during the `with` block is represented by a Session.
            
    Until better Session abstraction exists at the Python level,
    _session_resources may be added either at Context initialization or during
    session entry. _session_resources will be closed and removed at exit.
    """

    _session_ensemble_communicator: "Optional[mpi4py.MPI.Comm]" = None
    """Cached Session resource for communication between ensemble members.

    Retrieved from _session_resources before a Session for lower overhead
    in `ensemble_update()`.
    """

    _base_rank: Optional[int] = None
    """Cached rank in the parent allocation.

    We release our reference to the resource assignment when exiting the Session,
    but it is convenient to retain the rank that we were on, for debugging purposes.

    This attribute is not necessarily available before the Session is being entered,
    so it is initially None. 
    """

    def __init__(
        self,
        work=None,
        *args,
        resources: "Optional[ResourceAssignment[mpi4py.MPI.Comm, Any]]" = None,
        **kwargs,
    ):
        """Create manager for computing resources.

        Does not initialize resources because Python objects by themselves do
        not have a good way to deinitialize resources. Instead, resources are
        initialized using the Python context manager protocol when sessions are
        entered and exited.

        Appropriate computing resources need to be knowable when the Context is created.

        Keyword Arguments:
            work : work specification with which to initialize this context
            workdir_list (optional): deprecated
            communicator (optional): non-owning reference to a multiprocessing communicator
                from which ensemble and simulation session resources will be derived (deprecated)
            resources (optional): non-owning reference to resources (such as an MPI communicator)
                from which ensemble and simulator resources will be drawn. Caller is responsible
                for providing resources that are compatible with the work load.
                (Use instead of *communicator*.)

        .. versionchanged:: 0.4

            Keyword arguments after *work* are now keyword-only. Positional use
            is deprecated and will be unsupported in a future release for easier
            maintenance of the function signature.

        If provided, communicator must implement the mpi4py.MPI.Comm interface. The
        Context will use this communicator as the parent for subcommunicators
        used when launching sessions. If provided, communicator is owned by the
        caller, and must be freed by the caller after any sessions are closed.
        By default, the Context will get a reference to MPI_COMM_WORLD, which
        will be freed when the Python process ends and cleans up its resources.
        The communicator stored by the Context instance will not be used directly,
        but will be duplicated when launching sessions using ``with``.
        """

        # self.__context_array = list([Context(work_element) for work_element in work])
        from .workflow import WorkSpec
        from ..runtime import BaseContext

        # Handle positional arguments until we can require key-word-only.
        # TODO(release-2024): Simplify this logic after a suitable deprecation period to disallow *args.
        if not set(kwargs.keys()).issubset({"workdir_list", "communicator"}):
            raise TypeError("Unsupported key word arguments.")
        if (
            len(args) > 2
            or (len(args) == 2 and len(kwargs) > 0)
            or (len(args) == 1 and "workdir_list" in kwargs)
        ):
            raise TypeError("Unsupported positional arguments.")
        workdir_list = kwargs.get("workdir_list", None)
        communicator: Optional["mpi4py.MPI.Comm"] = kwargs.get("communicator", None)
        if len(args) >= 1:
            warnings.warn(
                "Deprecated use of positional argument. Use explicit key word.",
                DeprecationWarning,
            )
            workdir_list = args[0]
        if len(args) == 2:
            communicator = args[1]
        if communicator is not None:
            warnings.warn(
                "Provide *resources* instead of *communicator*.", DeprecationWarning
            )
        # End: handling for deprecated positional argument handling.

        if resources is not None and communicator is not None:
            raise exceptions.UsageError(
                "Do not provide both *communicator* and *resources*."
            )

        if resources is None:
            if communicator is not None:
                if BaseContext.communicator() is None:
                    BaseContext.set_base_communicator(communicator)
                elif BaseContext.communicator() != communicator:
                    raise exceptions.UsageError(
                        "Default base communicator already configured. "
                        "For non-trivial resource management, "
                        "provide Context with *resources* instead of *communicator*."
                    )
            self.__session_allocation = BaseContext.instance()
        else:
            # We do not intend to take ownership of the resources, but until we can
            # re-write the context manager as a function call (such as a `launch()`
            # method), we should not impose on the caller to maintain a reference to
            # the *resources* if the caller has no need for it and no responsibility
            # to clean it up. Hold a reference at least until the context manager exits.
            self._session_resources = resources

        self.__work = WorkSpec()
        self.__workdir_list = workdir_list

        self._session = None
        self.ensemble_rank = None
        self.simulator_rank = None

        # `work_width` notes the required width of an array of synchronous tasks
        # to perform the specified work.
        # As work elements are processed, self.work_width will be increased as appropriate.
        self.work_width = None

        # initialize the operations map. May be extended during the lifetime of a Context.
        # Note that there may be a difference between built-in operations provided by this module and
        # additional operations registered at run time.
        self.__operations = dict()
        # The map contains a builder for each operation. The builder is created
        # by passing the element to the function in the map. The object returned
        # must have the following methods:
        #
        #   * add_subscriber(another_builder) : allow other builders to subscribe
        #       to this one.
        #   * build(dag) : Fulfill the builder responsibilities by adding an
        #       arbitrary number of nodes and edges to a Graph.
        #
        # The gmxapi namespace of operations should be consistent with a
        # specified universal set of functionalities.
        self.__operations["gmxapi"] = {
            "md": lambda element: _md(self, element),
        }
        # Even if TPR file loading were to become a common and stable enough
        # operation to be specified in an API, it is unlikely to be implemented
        # by any code outside GROMACS, so let's not clutter a potentially more
        # universal namespace.
        self.__operations["gromacs"] = {
            "load_tpr": lambda element: _load_tpr(self, element),
        }

        # Right now we are treating workspec elements and work DAG nodes as equivalent, but they are
        # emphatically not intended to be tightly coupled. The work specification is intended to be
        # simple, user-friendly, general, and easy-to-implement. The DAG is an implementation detail
        # and may differ across context types. It is likely to have stronger typing of nodes and/or
        # edges. It is not yet specified whether we should translate the work into a graph before, after,
        # or during processing of the elements, so it is not yet known whether we will need facilities
        # to allow cross-referencing between the two graph-type structures. If we instantiate API objects
        # as we process work elements, and the DAG in a context deviates from the work specification
        # topology, we would need to use named dependencies to look up objects to bind to. Such facilities
        # could be hidden in the WorkElement class(es), too, to delegate code away from the Context as a
        # container class growing without bounds...
        # In actuality, we will have to process the entire workspec to some degree to make sure we can
        # run it on the available resources.
        self.elements = None

        # This setter must be called after the operations map has been populated.
        self.work = work

    @property
    def work(self):
        return self.__work

    @work.setter
    def work(self, work):
        """Set `work` attribute.

        Raises:
            gmx.exceptions.ApiError: work is not compatible with schema or
                known operations.
            gmx.exceptions.UsageError: Context can not access operations in
                the name space given for an Element
            gmx.exceptions.ValueError: assignment operation cannot be performed
                for the provided object (rhs)

        For discussion on error handling, see https://github.com/kassonlab/gmxapi/issues/125
        """
        from .workflow import WorkSpec, WorkElement

        if work is None:
            return

        if isinstance(work, WorkSpec):
            workspec = work
        elif hasattr(work, "workspec") and isinstance(work.workspec, WorkSpec):
            workspec = work.workspec
        else:
            raise exceptions.ValueError(
                "work argument must provide a gmx.workflow.WorkSpec."
            )
        workspec._context = self

        # Make sure this context knows how to run the specified work.
        for e in workspec.elements:
            element = WorkElement.deserialize(workspec.elements[e])

            # Note: Non-built-in namespaces (non-native) are treated as modules to import.
            # Native namespaces may not be completely implemented in a particular version of a particular Context.
            if element.namespace in {"gmxapi", "gromacs"}:
                assert element.namespace in self.__operations
                if not element.operation in self.__operations[element.namespace]:
                    # The requested element is a built-in operation but not available in this Context.
                    # element.namespace should be mapped, but not all operations are necessarily implemented.
                    logger.error(
                        "Operation {} not found in map {}".format(
                            element.operation, str(self.__operations)
                        )
                    )
                    # This check should be performed when deciding if the context is appropriate for the work.
                    # If we are just going to use a try/catch block for this test, then we should differentiate
                    # this exception from those raised due to incorrect usage.
                    # The exception thrown here may evolve with https://github.com/kassonlab/gmxapi/issues/125
                    raise exceptions.FeatureNotAvailableError(
                        "Specified work cannot be performed due to unimplemented operation {}.{}.".format(
                            element.namespace, element.operation
                        )
                    )

            else:
                assert element.namespace not in {"gmxapi", "gromacs"}

                # Don't leave an empty nested dictionary if we couldn't map the operation.
                if element.namespace in self.__operations:
                    namespace_map = self.__operations[element.namespace]
                else:
                    namespace_map = dict()

                # Set or update namespace map iff we have something new.
                if element.operation not in namespace_map:
                    try:
                        element_module = importlib.import_module(element.namespace)
                        element_operation = getattr(element_module, element.operation)
                    except ImportError as e:
                        raise exceptions.UsageError(
                            "Cannot find implementation for namespace {}. ImportError: {}".format(
                                element.namespace, str(e)
                            )
                        )
                    except AttributeError:
                        raise exceptions.UsageError(
                            "Cannot find factory for operation {}.{}".format(
                                element.namespace, element.operation
                            )
                        )
                    namespace_map[element.operation] = element_operation
                    self.__operations[element.namespace] = namespace_map

        self.__work = workspec

    def add_operation(self, namespace, operation, get_builder):
        # noinspection PyUnresolvedReferences
        """Add a builder factory to the operation map.

        Extends the known operations of the Context by mapping an operation in a namespace to a function
        that returns a builder to process a work element referencing the operation. Must be called before
        the work specification is added, since the spec is inspected to confirm that the Context can run it.

        It may be more appropriate to add this functionality to the Context constructor or as auxiliary
        information in the workspec, or to remove it entirely; it is straight-forward to just add snippets
        of code to additional files in the working directory and to make them available as modules for the
        Context to import.

        Example:

            >>> # Import some custom extension code.
            >>> import myplugin
            >>> myelement = myplugin.new_element()
            >>> workspec = gmx.workflow.WorkSpec()
            >>> workspec.add_element(myelement)
            >>> context = gmx.context.ParallelArrayContext()
            >>> context.add_operation(myelement.namespace, myelement.operation, myplugin.element_translator)
            >>> context.work = workspec
            >>> with get_context() as session:
            ...    session.run()

        """
        if namespace not in self.__operations:
            if namespace in {"gmxapi", "gromacs"}:
                raise exceptions.UsageError(
                    "Cannot add operations to built-in namespaces."
                )
            else:
                self.__operations[namespace] = dict()
        else:
            assert namespace in self.__operations

        if operation in self.__operations[namespace]:
            raise exceptions.UsageError(
                "Operation {}.{} already defined in this context.".format(
                    namespace, operation
                )
            )
        else:
            self.__operations[namespace][operation] = get_builder

    # Set up a simple ensemble resource
    # This should be implemented for Session, not Context, and use an appropriate subcommunicator
    # that is created and freed as the Session launches and exits.
    def ensemble_update(self, send, recv, tag=None):
        """Implement the ensemble_update member function that gmxapi through 0.0.6 expects.

        This is a draft of a Context feature that may not be available in all
        Context implementations.
        This feature requires that the Context is able to provide the
        ensemble_communicator feature and the numpy feature.
        """
        # gmxapi through 0.0.6 expects to bind to this member function during "build".
        # This behavior needs to be deprecated (bind during launch, instead), but this
        # dispatching function should be an effective placeholder.

        if tag is None or str(tag) == "":
            raise exceptions.ApiError("ensemble_update must be called with a name tag.")

        ensemble_communicator = self._session_ensemble_communicator

        assert not tag is None
        assert str(tag) != ""
        if tag not in self.part:
            self.part[tag] = 0
        logger.debug("Performing ensemble update.")
        ensemble_communicator.Allreduce(send, recv)
        buffer = np.array(recv, copy=False)
        buffer /= self.work_width
        suffix = "_{}.npz".format(tag)
        # These will end up in the working directory and each ensemble member will have one
        filename = str(
            "rank{}part{:04d}{}".format(
                ensemble_communicator.Get_rank(), int(self.part[tag]), suffix
            )
        )
        np.savez(filename, recv=recv)
        self.part[tag] += 1

    def __enter__(self):
        """Implement Python context manager protocol, producing a Session.

        The work specified in this Context is inspected to build a directed
        acyclic dependency graph (DAG). A Session is launched after reconciling
        the configured work with available computing resources. Each time the
        `run()` method of the Session is called, the graph is executed, and any
        nodes that have no more work to do are pruned from the graph.

        In practice, there are not currently any types of work implemented that
        do not complete on the first pass.

        Returns:
            Session object that can be run and/or inspected.

        Additional API operations are possible while the Session is active. When used as a Python context manager,
        the Context will close the Session at the end of the `with` block by calling `__exit__`.

        Note: this is probably where we will have to process the work specification to determine whether we
        have appropriate resources (such as sufficiently wide parallelism). Until we have a better Session
        abstraction, this means the clean approach should take two passes to first build a DAG and then
        instantiate objects to perform the work. In the first implementation, we kind of muddle things into
        a single pass.

        TODO(#3145, #4079): Use an explicit function call for context manager behavior.
        By implementing __enter__ and __exit__ directly on the class, we prevent
        ourselves from cleanly nesting context managers for managed resources,
        and we limit our ability to decouple the code involved in setting up and
        tearing down a Session.
        """
        # Cache the working directory from which we were launched
        # so that __exit__() can give us proper context management behavior.
        self.__initial_cwd = os.getcwd()
        from ..runtime import ResourceAssignment, ResourceRequirements

        try:
            import networkx as nx
            from networkx import DiGraph as _Graph
        except ImportError:
            raise exceptions.FeatureNotAvailableError(
                "gmx requires the networkx package to execute work graphs."
            )

        logger.debug("Launching session from {}".format(self.__initial_cwd))

        if self._session is not None:
            raise exceptions.Error("Already running.")
        if self.work is None:
            raise exceptions.UsageError("No work to perform!")

        # Set up the global and local context.
        # Check the global MPI configuration
        logger.debug(
            f'Starting session for GROMACS MPI type "{gmxapi.utility.config()["gmx_mpi_type"]}"'
        )

        ###
        # Process the work specification.
        ###
        logger.debug("Processing workspec:\n{}".format(str(self.work)))

        # Get a builder for DAG components for each element
        builders = {}
        builder_sequence = []
        for element in self.work:
            # dispatch builders for operation implementations
            try:
                new_builder = self.__operations[element.namespace][element.operation](
                    element
                )
                assert hasattr(new_builder, "add_subscriber")
                assert hasattr(new_builder, "build")

                logger.info("Collected builder for {}".format(element.name))
            except LookupError as e:
                request = ".".join([element.namespace, element.operation])
                message = "Could not find an implementation for the specified operation: {}. ".format(
                    request
                )
                message += str(e)
                raise exceptions.ApiError(message)
            # Subscribing builders is the Context's responsibility because otherwise the builders
            # don't know about each other. Builders should not depend on the Context unless they
            # are a facility provided by the Context, in which case they may be member functions
            # of the Context. We will probably need to pass at least some
            # of the Session to the `launch()` method, though...
            dependencies = element.depends
            for dependency in dependencies:
                # If a dependency is a list, assume it is an "ensemble" of dependencies
                # and pick the element corresponding to the local rank.
                if isinstance(dependency, (list, tuple)):
                    if self._session_resources is None:
                        raise exceptions.FeatureNotAvailableError(
                            "Cannot infer dependency graph topology. Provide a ResourceAssignment instead "
                            "of a raw Communicator for complex dependencies."
                        )
                    _member_id = self._session_resources.subtask_id()
                    if _member_id is not None:
                        assert len(dependency) > _member_id
                        name = str(dependency[_member_id])
                        del _member_id
                    else:
                        name = None
                else:
                    name = dependency
                if name is not None:
                    logger.info("Subscribing {} to {}.".format(element.name, name))
                    builders[name].add_subscriber(new_builder)
            builders[element.name] = new_builder
            builder_sequence.append(element.name)
            logger.debug("Appended {} to builder sequence.".format(element.name))

        # Call the builders in dependency order
        # Note: session_communicator is available, but ensemble_communicator has not been created yet.
        graph = _Graph(width=1)
        logger.info("Building sequence {}".format(builder_sequence))
        for name in builder_sequence:
            builder = builders[name]
            logger.info("Building {}".format(builder))
            logger.debug("Has build attribute {}.".format(builder.build))
            builder.build(graph)
            logger.debug("Built.")
        self.work_width = graph.graph["width"]

        # Support the legacy invocation (received *communicator* instead of *resources*)
        if self.__session_allocation is not None:
            assert self._session_resources is None
            # We will need to own this resource assignment, being sure to close
            # and release in __exit__.
            _ensemble_label = 0
            requirements = ResourceRequirements(
                communication=tuple(
                    {"ensemble_comm": _ensemble_label, "subtask_comm": True}
                    for _ in range(self.work_width)
                )
            )
            self._session_resources = ResourceAssignment.assign(
                allocation=self.__session_allocation, requirements=requirements
            )

        self.ensemble_rank = self._session_resources.subtask_id()
        self.simulator_rank = self._session_resources.subtask_rank()
        self._base_rank = self._session_resources.base_rank()
        member_id = self.ensemble_rank

        assert self._session_resources is not None

        ensemble_comm = self._session_resources.ensemble_communicator()
        if ensemble_comm:
            _ensemble_size = ensemble_comm.Get_size()
            if self.work_width != _ensemble_size:
                raise ProtocolError(
                    f"Resource mismatch: {self.work_width} subtasks received ensemble resources for "
                    f"{_ensemble_size}."
                )
            assert ensemble_comm.Get_rank() == self.ensemble_rank
            self._session_ensemble_communicator = ensemble_comm

        # Provide a Session-scoped gmxapi._gmxapi.Context.
        # Provide an alternative to MPI_COMM_WORLD for the SimulatorCommunicator.
        # libgromacs does not accept a null communicator at run time, but our calling
        # convention is to make the same API calls on all ranks at the WORLD level
        # (or at least at the Context.__communicator level).
        # self._api_context needs to be in place before the launch() functor from
        # the _md Builder is called.
        simulation_communicator = self._session_resources.communicator()

        if simulation_communicator:
            # Prepare working directories.
            # This should probably be moved to some aspect of the Session and either
            # removed from here or made more explicit to the user.
            workdir_list = self.__workdir_list
            if workdir_list is None:
                workdir_list = [
                    os.path.join(".", str(i)) for i in range(self.work_width)
                ]
            workdir_list = [os.path.abspath(directory) for directory in workdir_list]

            self.workdir = simulation_communicator.bcast(workdir_list[member_id])
            del workdir_list

            if self._session_resources.is_subtask_root() and not os.path.exists(
                self.workdir
            ):
                os.mkdir(self.workdir)

            try:
                self._api_context = _gmxapi.create_context(simulation_communicator)
            except Exception as e:
                logger.error(
                    "Got exception when trying to create default library context.",
                    exc_info=e,
                )
                # Warning: Be careful about suppressing exceptions.
                #
                # Avoid collective calls in `finally` blocks or context manager
                # `__exit__` functions!
                #
                # Suppressing exceptions could cause hangs on errors.
                # There is potential divergence in collective MPI calls
                # if all ranks do not proceed as expected.
                # If another rank may be in (or approaching) a collective call,
                # we must exit with an uncaught exception (or explicitly
                # MPI_ABORT). If an exception may cause divergence between ranks,
                # we must resolve it (or allow MPI_ABORT) before entering another
                # collective call.
                #
                # Here, the exception raised is likely more
                # informative than the AssertionError that would occur below,
                # so re-raise.
                #
                # For the present context manager, the collective calls in __exit__
                # are okay because __exit__ will not be called if __enter__ does
                # not complete successfully.
                raise e

        if simulation_communicator:
            assert self._api_context is not None
        else:
            assert self._api_context is None

        # In the first implementation, we only provide the ensemble
        # communicator to the simulation master ranks (as we do for tMPI).
        # In the long term, it may be appropriate for all simulator ranks to
        # have access to the ensemble communicator, but that would be tricky
        # to get right. This requires that the work load manager (the current
        # code block) needs to know (or be able to defer negotiation of)
        # something about the library-internal resource allocation.
        # (See also https://gitlab.com/gromacs/gromacs/-/issues/3718)
        #
        # Also note: The GROMACS core does not have a way to receive a
        # communicator that is consistent with the notion of the "ensemble communicator".
        # The `_session_ensemble_communicator` is currently only a gmxapi-specific
        # Session resource usable by MD extension code based on sample_restraint.
        # We should reconsider this in the context of current and future use of
        # the multisim communicator.
        self.part = {}

        # launch() is currently a method of gmx.core.MDSystem and returns a gmxapi::Session.
        # MDSystem objects are obtained from gmx.core.from_tpr(). They also provide add_potential().
        # gmxapi::Session objects are exposed as gmx.core.MDSession and provide run() and close() methods.
        #
        # Here, I want to find the input appropriate for this rank and get an MDSession for it.
        # E.g. Make a pass that allows meta-objects to bind (setting md_proxy._input_tpr and md_proxy._plugins,
        # and then call a routine implemented by each object to run whatever protocol it needs, such
        # as `system = gmx.core.from_tpr(md._input_tpr); system.add_potential(md._plugins)
        # For future design plans, reference https://github.com/kassonlab/gmxapi/issues/65
        #
        # This `if` condition is currently the thing that ultimately determines whether the
        # rank attempts to do work.
        if simulation_communicator:
            # print(graph)
            logger.debug(("Launching graph {}.".format(graph.graph)))
            logger.debug("Graph nodes: {}".format(str(list(graph.nodes))))
            logger.debug("Graph edges: {}".format(str(list(graph.edges))))

            simulator_rank = simulation_communicator.Get_rank()
            logger.info(
                "Launching work for member {}, subcommunicator rank {}.".format(
                    member_id, simulator_rank
                )
            )

            # Launch the work for this rank
            simulation_communicator.barrier()
            if not os.path.isdir(self.workdir):
                raise exceptions.Error(
                    "{} is not a valid working directory.".format(self.workdir)
                )
            os.chdir(self.workdir)
            logger.info(
                "rank {} changed directory to {}".format(member_id, self.workdir)
            )
            sorted_nodes = nx.topological_sort(graph)
            runners = []
            closers = []
            for name in sorted_nodes:
                launcher = graph.nodes[name]["launch"]
                # TODO(#4079): launch is explicitly a collaboration with the Context
                #  to build a Session, which should provide SessionResources here.
                runner = launcher(member_id)
                if runner is not None:
                    runners.append(runner)
                    closers.append(graph.nodes[name]["close"])

            # Get a session object to return. It must simply provide a `run()` function.
            class Session(object):
                _context: _gmxapi.Context = None

                def __init__(self, runners, closers, library_context: _gmxapi.Context):
                    self.runners = list(runners)
                    self.closers = list(closers)
                    if library_context is None:
                        raise ProtocolError(
                            "Simulation Session requires a Context from the C++ extension module."
                        )
                    self._context = library_context

                def run(self):
                    # Note we are not following the documented protocol of running repeatedly yet.
                    to_be_deleted = []
                    for i, runner in enumerate(self.runners):
                        try:
                            runner()
                        except StopIteration:
                            to_be_deleted.insert(0, i)
                    for i in to_be_deleted:
                        del self.runners[i]
                    return True

                def close(self):
                    for close in self.closers:
                        logger.debug("Closing node: {}".format(close))
                        close()
                    del self._context

            self._session = Session(runners, closers, self._api_context)
        else:
            logger.warning("Process {} has no work to do".format(os.getpid()))

            class NullSession:
                def __init__(self, base_rank: int):
                    self.rank = base_rank

                def run(self):
                    logger.info(f"Running null session on rank {self.rank}.")
                    return True

                def close(self):
                    logger.info("Closing null session.")

            assert self._api_context is None
            self._session = NullSession(self._session_resources.base_rank())

        # Make sure session has started on all ranks before continuing?

        self._session.graph = graph
        return self._session

    def __exit__(self, exception_type, exception_value, traceback):
        """Implement Python context manager protocol."""
        try:
            logger.info(
                "Exiting session on rank {}.".format(
                    self._session_resources.base_rank()
                )
            )
            if self._session is not None:
                logger.info("Calling session.close().")
                self._session.close()
            else:
                # Note: we should not have a None session but rather an API-compliant Session that just has no work.
                # Reference: https://github.com/kassonlab/gmxapi/issues/41
                logger.info("No _session known to context or session already closed.")

            if self._api_context is not None:
                # It really seems like we should expect an explicit close() method in the future.
                if hasattr(self._api_context, "close"):
                    try:
                        self._api_context.close()
                    except _gmxapi.Exception:
                        logger.exception("Error while closing GROMACS API context.")
                logger.debug(f"De-referencing library context {self._api_context}")
                self._api_context = None

            # We only need to free resources if we allocated them here because the caller
            # did not provide session resources.
            if self.__session_allocation is not None:
                # Support the legacy invocation (*communicator* instead of *resources*)
                if self._session_resources is not None:
                    logger.debug("Freeing session resources.")
                    self._session_resources.close()
        finally:
            self._session_resources = None
            if self.__session_allocation is not None:
                logger.debug("Releasing session allocation.")
                self.__session_allocation = None
            os.chdir(self.__initial_cwd)

        self._session = None
        self._session_ensemble_communicator = None
        self.simulator_rank = None
        # Leave `ensemble_rank` set to help clients find the right results afterwards.
        # I.e. do not set `self.ensemble_rank = None` here.

        logger.info("Session closed on context rank {}.".format(self._base_rank))
        # Note: Since sessions running in different processes can have different
        # work, sessions have not necessarily ended on all ranks. As a result,
        # starting another session on the same resources could block until the
        # resources are available.

        # Python context managers return False when there were no exceptions to handle.
        return False


def get_context(work=None):
    """Get a concrete Context object.

    Args:
        work (gmx.workflow.WorkSpec): runnable work as a valid gmx.workflow.WorkSpec object

    Returns:
        An object implementing the :py:class:`gmx.context.Context` interface, if possible.

    Raises:
        gmx.exceptions.ValueError if an appropriate context for ``work`` could not be loaded.

    If work is provided, return a Context object capable of running the provided work or produce an error.

    The semantics for finding Context implementations needs more consideration, and a more informative exception
    is likely possible.

    A Context can run the provided work if

      * the Context supports can resolve all operations specified in the elements
      * the Context supports DAG topologies implied by the network of dependencies
      * the Context supports features required by the elements with the specified parameters,
        such as synchronous array jobs.

    """
    # We need to define an interface for WorkSpec objects so that we don't need
    # to rely on typing and inter-module dependencies.
    from .workflow import WorkSpec

    workspec = None
    if work is not None:
        if isinstance(work, WorkSpec):
            workspec = work
        elif hasattr(work, "workspec") and isinstance(work.workspec, WorkSpec):
            workspec = work.workspec
        else:
            raise exceptions.ValueError(
                "work argument must provide a gmx.workflow.WorkSpec."
            )
    if (
        workspec is not None
        and hasattr(workspec, "_context")
        and workspec._context is not None
    ):
        context = workspec._context
    else:
        context = Context(work=workspec)

    return context



#
# This file is part of the GROMACS molecular simulation package.
#
# Copyright 2019- The GROMACS Authors
# and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
# Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
#
# GROMACS is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# GROMACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with GROMACS; if not, see
# https://www.gnu.org/licenses, or write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
#
# If you want to redistribute modifications to GROMACS, please
# consider that scientific software is very special. Version
# control is crucial - bugs must be traceable. We will be happy to
# consider code for inclusion in the official distribution, but
# derived work must not be called official GROMACS. Details are found
# in the README & COPYING files - if they are missing, get the
# official version at https://www.gromacs.org.
#
# To help us fund GROMACS development, we humbly ask that you cite
# the research papers on the package. Check out https://www.gromacs.org.

"""Abstract base classes for gmxapi Python interfaces.

This module consolidates definitions of some basic interfaces in the gmxapi
Python package. These definitions are evolving and mostly for internal use, but
can be used to check compatibility with the gmxapi implementation details that
are not otherwise fully specified by the API.

Refer to [PEP 484](https://www.python.org/dev/peps/pep-0484) and to
[PEP 526](https://www.python.org/dev/peps/pep-0526) for Python background.

Developer Note:
    Type checking fails when accessing attributes of generic classes (see PEP 484).
    This affects some scoping choices.

Developer Note:
    It is worth noting some details regarding type erasure with Python generics.
    Generic type parameters (class subscripts) allow for annotation of dynamic
    types that are bound when the generic is instantiated. The subscript is not
    part of the run time class. This means that we can use subscripted generic
    classes for static type checking, but use for any other purpose is discouraged.
    Thus, type variable parameters to generics have less meaning that do template
    parameters in C++, and are orthogonal to subclassing. Note, though, that a
    subclass is not generic if the generic parameters are bound (explicitly
    with type subscripts or implicitly by omission (implying `typing.Any`).

    Also, type parameters
    (or types of parameters) to functions can be used to dispatch a generic
    function to a specific function.

    In other words: keep in mind the
    orthogonality of generic classes and base classes, and recognize that
    composed objects mimicking C++ template specializations are not distinguishable
    at the class level.

Note: This module overly specifies the API. As we figure out the relationships.
    As we clarify interactions, we should trim these specifications or migrate them
    to gmxapi.typing and focus on developing to functions and function interfaces
    rather than classes or types. In the mean time, these types help to illustrate
    the entities that can exist in the implementation. Just keep in mind that
    these ABCs describe the existence of "protocols" more than the details of
    the protocols, which may take longer to define.

..  todo:: Clarify protocol checks, ABCs, and mix-ins.

"""
# Note that the Python typing module defines generic classes in terms of abstract
# base classes defined in other modules (namely `collections`), but without
# actual inheritance. The ABCs are not intended to be instantiated, and the
# generics are very mangled objects that cannot be instantiated. However,
# user-defined subclasses of either may be instantiated.
#
# This restriction is probably due somewhat to implementation constraints in the
# typing module, but it represents a Separation of Concerns that we should
# consider borrowing in our model. Practically, this can mean using an abstract
# for run time checking and a generic for static type checking.
#
# There may not be a compelling reason to
# rigorously separate ABC and generic type, or to disallow instantiating a
# generic unless it is also an abstract. Note the distinction, though, between
# abstract generics and fully implemented generics. NDArray
# and Future are likely to be examples of fully implemented generics, while
# Context and various interface types are likely to have abstract generics.
#
# Use abstract base classes to define interfaces and relationships between
# interfaces. Use `typing` module machinery for static type checking. Use the
# presence or absence of an expected interface, or exception handling, for run
# time type checking. Use `isinstance` and `issubclass` checking against
# non-generic abstract base classes when the valid interface is complicated or
# unknown in the caller's context.
#
# Also, note that abstract base classes cannot assert that a simple data member
# is provided by a subclass, but class data members of Generic classes are
# interpreted by the type checker as instance members (unless annotated ClassVar).

import typing
from abc import ABC, abstractmethod
import collections.abc
from typing import Type, Callable


class NDArray(collections.abc.Sequence, ABC):
    """N-dimensional data interface."""

    # TODO: Fix the data model and ABC vs. type-checking conflicts so that we can
    #       recognize NDArrays and NDArray futures as compatible data in data flow operations.
    # We cannot do the following because of the forward reference to Future:
    # @classmethod
    # def __subclasshook__(cls, subclass):
    #     """Determine whether gmxapi should consider the provided type to be consistent with an NDArray."""
    #     if subclass is cls:
    #         return True
    #     if cls is NDArray:
    #         # For the purposes of gmxapi data flow, a Future[NDArray] is equivalent to an NDArray.
    #         if Future in subclass.__mro__:
    #             return issubclass(subclass.dtype, cls)
    #         else:
    #             def is_compatible(candidate):
    #                 if issubclass(candidate, collections.abc.Sequence) \
    #                         and not issubclass(candidate, (str, bytes)):
    #                     return True
    #                 return False
    #             any(is_compatible(base) for base in subclass.__mro__)
    #             return True
    #     return NotImplemented


NDArray.register(list)


# TODO: Define an enumeration.
SourceProtocol = typing.NewType("SourceProtocol", str)


class Resource(ABC):
    """gmxapi resource object interface.

    Resources are generally provided by Operation instances, but may be provided
    by the framework, or abstractly as the outputs of other scripts.

    A resource is owned in a specific Context. A resource handle may be created
    in a different Context than that in which the resource is owned. The Context
    instances negotiate the resource availability in the new Context.
    """

    # No public interface is yet defined for the Resource API.

    # Resource instances should have an attribute serving as a sequence of available
    # source protocols in decreasing order of preference.
    # TODO: Clarify. Define an enumeration with the allowed values.
    # TODO: Enforce. The existence of this attribute cannot be confirmed by the abc.ABC machinery.
    # TODO: Consider leaving this off of the ABC specification and make it an implementation
    #  detail of a single_dispatch function.
    # TODO: Similarly, find a way for the subscriber to have a chain of handlers.
    _gmxapi_source_protocol = typing.Sequence[SourceProtocol]


class Future(Resource):
    """Data source that may represent Operation output that does not yet exist.

    Futures represent "immutable resources," or fixed points in the data flow.
    """

    @property
    @abstractmethod
    def dtype(self) -> type:
        """Data type for the promised result."""
        ...

    @abstractmethod
    def result(self) -> typing.Any:
        """Fetch data to the caller's Context.

        Returns the actual result of the operation supplying this Future.

        Ensemble data are returned as a list. Scalar results or results from single
        member ensembles are returned as scalars. If this behavior is confusing or
        problematic for you, please reopen https://gitlab.com/gromacs/gromacs/-/issues/3179
        or a new issue and join the discussion.
        """
        ...

    @classmethod
    def __subclasshook__(cls, C):
        if cls is Future:
            if any("result" in B.__dict__ and callable(B.result) for B in C.__mro__):
                return True
        return NotImplemented


class MutableResourceSubscriber(ABC):
    """Required interface for subscriber of a MutableResource.

    A MutableResource is bound to a collaborating object by passing a valid
    Subscriber to the resource's *subscribe()* method.
    """


class MutableResource(Resource):
    """An Operation interface that does not represent a fixed point in a data flow.

    Providers and consumers of mutable resources are more tightly coupled than
    data edge terminals and have additional machinery for binding at run time.
    Examples include the simulation plugin interface, binary payloads outside of
    the standard gmxapi types, and any operation interaction that the current
    context must defer to lower-level details.

    There is not yet a normative interface for MutableResources, but a consumer
    of MutableResources has chance to bind directly to the provider of a
    MutableResource without the mediation of a DataEdge. Accordingly, providers
    and consumers of MutableResources must be able to be instantiated in the
    same Context.
    """

    def subscribe(self, subscriber: MutableResourceSubscriber):
        """Create a dependency on this resource.

        Allows a gmxapi compatible operation to bind to this resource.
        The subscribing object will be provided with a lower-level registration
        interface at run time, as computing elements are being initialized.
        The nature of this registration interface is particular to the type of
        resource and its participants. See, for instance, the MD plugin binding
        interface.
        """


class OutputDataProxy(ABC):
    """A collection of Operation outputs.

    This abstract base class describes the interface to the output of an
    operation / work node.
    """

    # TODO: Specification.
    # Currently, the common aspect of OutputDataProxy is that a class has public
    # attributes that are exclusively OutputDescriptor instances, meaning that
    # getattr(instance, attr) returns a Future object. However, there are several
    # ways to implement getattr, and this sort of check in an ABC does not appear
    # to be common.
    #
    # We might choose to assert that all public attributes must be compatible data
    # descriptors, in conjunction with defining a more specific OutputDataProxy
    # metaclass, but this makes for a dubiously growing chain of data descriptors
    # we use for the output access.
    #
    # The data model might be cleaner if we move to something more
    # like a Collection or Mapping with more conventional getters, but we would
    # lose the ability to check type on individual elements. (However, the
    # typing of return values is not normally the defining aspect of an ABC.)
    #
    # Another alternative is to collapse the contents of the `output` attribute
    # into the Operation handle type, strongly define all handle types (so that
    # the type checker can identify the presence of attributes), and rely only
    # on type checking at the level of the data descriptors. (Dynamically defined
    # OutputDataProxy classes are the execption, rather than the rule.)
    #
    # We will need to consider the details of type checkers and syntax inspection
    # tools, like Jedi and MyPy, to make design choices that maximize API usability
    # and discoverability.


class OperationReference(ABC):
    """Client interface to an element of computational work already configured.

    An "instance" of an operation is assumed to be a node in a computational
    work graph, owned and managed by a Context. This class describes the
    interface of the reference held by a client once the node exists.

    The convergence of OperationReferences with Nodes as the results of the action
    of a Director implies that a Python user should also be able to "subscribe"
    to an operation handle (or its member resources). This could be a handy feature
    with which a user could register a call-back. Note that we will want to provide
    an optional way for the call-back (as with any subscriber) to assert a chain
    of prioritized Contexts to find the optimal path of subscription.
    """

    @abstractmethod
    def run(self):
        """Assert execution of an operation.

        After calling run(), the operation results are guaranteed to be available
        in the local context.
        """

    @property
    @abstractmethod
    def output(self) -> OutputDataProxy:
        """Get a proxy collection to the output of the operation.

        Developer note: The 'output' property exists to isolate the namespace of
        output data from other operation handle attributes and we should consider
        whether it is actually necessary or helpful. To facilitate its possible
        future removal, do not enrich its interface beyond that of a collection
        of OutputDescriptor attributes. The OutputDataProxy also serves as a Mapping,
        with keys matching the attributes. We may choose to keep only this aspect
        of the interface instead of trying to keep track of the set of attributes.
        """
        ...


class Fingerprint(ABC):
    """Unique global identifier for an Operation node.

    Represents the operation and operation inputs.

    No public interface.
    """


class OutputDescription(ABC):
    """

    There may not be a single OutputDescription base class, since the requirements
    are related to the Context implementation.
    """


class Edge(ABC):
    """Reference to the state and description of a data flow edge.

    A DataEdge connects a data source collection to a data sink. A sink is an
    input or collection of inputs of an operation (or fused operation). An operation's
    inputs may be fed from multiple data source collections, but an operation
    cannot be fully instantiated until all of its inputs are bound, so the DataEdge
    is instantiated at the same time the operation is instantiated because the
    required topology of a graph edge may be determined by the required topology
    of another graph edge.

    A data edge has a well-defined topology only when it is terminated by both
    a source and sink. Creation requires that a source collection is compared to
    a sink description.

    Calling code initiates edge creation by passing well-described data sources
    to an operation factory. The data sources may be annotated with explicit scatter
    or gather commands.

    The resource manager for the new operation determines the
    required shape of the sink to handle all of the offered input.

    Broadcasting
    and transformations of the data sources are then determined and the edge is
    established.

    At that point, the fingerprint of the input data at each operation
    becomes available to the resource manager for the operation. The fingerprint
    has sufficient information for the resource manager of the operation to
    request and receive data through the execution context.

    Instantiating operations and data edges implicitly involves collaboration with
    a Context instance. The state of a given Context or the availability of a
    default Context through a module function may affect the ability to instantiate
    an operation or edge. In other words, behavior may be different for connections
    being made in the scripting environment versus the running Session, and implementation
    details can determine whether or not new operations or data flow can occur in
    different code environments.

    A concrete Edge is a likely related to the consuming Context, and a single
    abstract base class may not be possible or appropriate. Possible Context-agnostic
    use cases for a global abstract Edge (along with Node) include topological aspects of data
    flow graphs or user-friendly inspection.
    """


class Node(ABC):
    """Object oriented interface to nodes configured in a Context.

    In gmxapi.operation Contexts, this functionality is implemented by subclasses
    of SourceResource.

    Likely additional interfaces for Node include subscribers(), label(), and
    (weak) reference helpers like context() and identifier().

    .. todo:: Converge. A Node is to a concrete Context what an operation handle is to the None (Python UI) Context.
    """

    @abstractmethod
    def handle(self, context: "Context") -> OperationReference:
        """Get a reference to the Operation in the indicated Context.

        This is equivalent to the reference obtained from the helper function
        or factory that added the node if and only if the Contexts are the same.
        Otherwise, a new node is created in *context* that subscribes to the
        original node.
        """
        # Note that a member function like this is the same as dispatching a
        # director that translates from the Node's Context to *context*

    @abstractmethod
    def output_description(self, context: "Context") -> OutputDescription:
        """Get a description of the output available from this node.

        Returns a subset of the information available through handle(), but
        without creating a subscription relationship. Allows data sources and
        consumers to determine compatibility and requirements for connecting
        two nodes.
        """

    @abstractmethod
    def input(self) -> Edge:
        """Describe the bound data sources.

        The returned object represents the data edge in the Context managing
        the node, though the data sources may be from other Contexts.
        """

    @abstractmethod
    def fingerprint(self) -> Fingerprint:
        """Uniquely identify this Node.

        Used internally to manage resources, check-point recovery, and messaging
        between Contexts. The fingerprint is dependent on the operation and the
        operation inputs, and is independent of the Context.

        Returns:
            Opaque identifier describing the unique output of this node.
        """

    @abstractmethod
    def operation(self) -> "OperationImplementation":
        """Get a reference to the registered operation that produces nodes like this."""
        # Note that the uniqueness of a node is such that Node.operation() and
        # Node.input() used to get an OperationReference in the same Context
        # should result in a handle to the same Node.


class NodeBuilder(ABC):
    """Add an element of computational work to be managed by a gmxapi Context.

    A Node generally represents an instance of a registered Operation, but the
    only real requirement is that it contains sufficient information to run an
    operation and to direct the instantiation of an equivalent node in a
    different consumer Context.

    In the near future, Node and NodeBuilder will participate in check-pointing
    and in a serialization/deserialization scheme.

    .. todo:: As the NodeBuilder interface is minimized, we can look for a normative
              way to initialize a Generic NodeBuilder that supports the sorts of
              type inference and hinting we would like.
    """

    @abstractmethod
    def build(self) -> OperationReference:
        """Finalize the creation of the operation instance and get a reference."""
        ...

    @abstractmethod
    def set_input_description(self, input_description):
        """Add the details related to the operation input.

        Example: In gmxapi.operation, includes signature() and make_uid()

        .. todo:: This can probably be moved to an aspect of the resource factory.
        """
        ...

    @abstractmethod
    def set_output_factory(self, output_factory):
        """Set the factory that gives output description and resources for the Node.

        Output is not fully describable until the input is known and the Node is
        ready to be instantiated. This is the resource that can be used by the
        Context to finish completely describing the Node. The interface of the
        factory and of any object it produces is a lower level detail of the
        Context and Operation implementations in that Context.

        .. todo:: This can probably be moved to an aspect of the resource factory.
        """

    @abstractmethod
    def add_input(self, name: str, source):
        """Attach a client-provided data source to the named input.

        .. todo:: Generalize to add_resource().
        """
        ...

    @abstractmethod
    def set_handle(self, handle_builder):
        """Set the factory that gives a builder for a handle to the operation.

        .. todo:: Stabilize interface to handle_builder and move to an aspect of the
                  operation registrant.
        """

    @abstractmethod
    def set_runner_director(self, runner_builder):
        """Set the factory that gives a builder for the run-time callable.

        .. todo:: This should be a specialized Director obtained from the registrant
                  by the Context when translating a Node for execution.
        """

    @abstractmethod
    def set_resource_factory(self, factory: Callable):
        """Register a resource factory for the operation run-time resources.

        The factory will be called within the Context

        .. todo:: Along with the merged set_runner/Director, the resource factory
                  is the other core aspect of an operation implementation registrant
                  that the Context should fetch rather than receiving through the
                  NodeBuilder.
        """
        # The factory function takes input in the form the Context will provide it
        # and produces a resource object that will be passed to the callable that
        # implements the operation.
        assert callable(factory)
        ...


class Context(ABC):
    """API Context.

    A Context instance manages the details of the computing environment and
    provides for the allocation of resources.
    All gmxapi data and operations are owned by a Context instance.
    The Context manages the details of how work is run and how data is managed.

    Additionally, a concrete Context implementation determines some details of
    the interfaces used to manage operation execution and data flow. Thus, API
    calls may depend on multiple Contexts when, for instance, there is a source
    Context, a caller Context, and/or a destination Context. For Python data
    types and external interfaces to the gmxapi package (such as public function
    signatures) is equal to *None*.

    This abstract base class (ABC) defines the required interface of a Context
    implementation. Client code should use this ABC for type hints. Concrete
    implementations may, *but are not required*, to subclass from this ABC to
    help enforce compatibility.
    """

    @abstractmethod
    def node_builder(self, *, operation, label=None) -> NodeBuilder:
        """Get a builder for a new work graph node.

        Nodes are elements of computational work, with resources and execution
        managed by the Context. The Context handles parallelism resources, data
        placement, work scheduling, and data flow / execution dependencies.

        This method is used by Operation director code and helper functions to
        add work to the graph.
        """
        ...

    @abstractmethod
    def node(self, node_id) -> Node:
        """Get the indicated node from the Context.

        node_id may be an opaque identifier or a label used when the node was
        added.
        """


class ResourceFactory(ABC):
    """Packager for run time resources for a particular Operation in a particular Context.

    TODO: refine interface.
    """

    @abstractmethod
    def input_description(self, context: Context):
        """Get an input description in a form usable by the indicated Context."""


class OperationDirector(ABC):
    """Interface for Operation entry points.

    An operation director is instantiated for a specific operation and context
    (by a dispatching factory) to update the work managed by the context
    (add a computational element).
    """

    # TODO: How to handle subscriptions? Should the subscription itself be represented
    #  as a type of resource that is passed to __call__, or should the director be
    #  able to subscribe to resources as an alternative to passing with __call__?
    #  Probably the subscription is represented by a Future passed to __call__.

    # TODO: Annotate `resources`, whose validity is determined by both context and operation.
    @abstractmethod
    def __call__(self, resources, label: typing.Optional[str]):
        """Add an element of work (node) and return a handle to the client.

        Implements the client behavior in terms of the NodeBuilder interface
        for a NodeBuilder in the target Context. Return a handle to the resulting
        operation instance (node) that may be specialized to provide additional
        interface particular to the operation.
        """
        ...

    @abstractmethod
    def handle_type(self, context: Context) -> Type[OperationReference]:
        """Get the class used for operation references in this Context.

        Convenience function. May not be needed.
        """
        ...

    @abstractmethod
    def resource_factory(
        self,
        source: typing.Union[Context, None],
        target: typing.Optional[Context] = None,
    ) -> typing.Union[ResourceFactory, typing.Callable]:
        """Get an appropriate resource factory.

        The ResourceFactory converts resources (in the form produced by the *source* Context)
        to the form consumed by the operation in the *target* Context.

        A *source* of None indicates that the source is an arbitrary Python function
        signature, or to try to detect appropriate dispatching. A *target* of
        None indicates that the Context of the Director instance is the target.

        As we merge the interface for a NodeBuilder and a RunnerBuilder, this will not need to be
        specified in multiple places, but it is not yet clear where the responsibility lies.
        Ultimately, the operation implementation should only need to provide a single-dispatching
        helper for finalizing input preparation in a target Context when building a node.
        Converting resources from one Context to another during edge creation is a
        collaboration between the resource type, the providing Context, and the consuming
        Context. It should be mediated by functions registered with the Context when the
        Operation is registered, and such functionality should be decoupled from the
        OperationDirector.

        Generally, the client should not need to call the resource_factory directly.
        The director should dispatch an appropriate factory. C++ versions need
        the resource_factory dispatcher to be available for reasons of compilation
        dependencies. Clients may want to use a specific resource_factory to explicitly
        control when and where data flow is resolved. Also, the resource_factory
        for the None Context can be used to set the function signature of the Python
        package helper function. As such, it may be appropriate to make it a "static"
        member function in Python.
        """
        ...


class OperationImplementation(ABC):
    """Essential interface of an Operation implementation.

    Describe the essential features of an Operation that can be registered with
    gmxapi to support building and executing work graphs in gmxapi compatible
    execution contexts.

    An Operation is usable in gmxapi when an OperationImplementation is registered
    with a valid identifier, consisting of a *namespace* and a *name*.

    Generally, the *namespace* is the module implementing the Operation
    and the *name* is a factory or helper importable from the same module that
    triggers the OperationDirector to configure a new instance of the Operation.

    The registered object must be able to describe its namespace and name, and
    to dispatch an OperationDirector appropriate for a given Context.
    """

    # TODO: Either OperationImplementations should be composed or subclasses should
    #  each be singletons. We can still consider that OperationReferences are instances
    #  of OperationImplementations or its subclasses, or that OperationReference classes
    #  have a class data member pointing to a single OperationImplementation instance
    #  or operation_registry value.

    # TODO: Consider a data descriptor and metaclass to validate the name and namespace.
    @classmethod
    @abstractmethod
    def name(self) -> str:
        """The name of the operation.

        Generally, this corresponds to a callable attribute of a Python module
        (named by namespace()) that acts as a factory for new operation instances.
        It is also used by Context implementations to locate code supporting
        the operation.
        """
        ...

    # TODO: Consider a data descriptor and metaclass to validate the name and namespace.
    @classmethod
    @abstractmethod
    def namespace(self) -> str:
        """The namespace of the operation.

        Generally, the namespace corresponds to a Python module importable in
        the execution environment.
        """

    # TODO: Allow this to be an instance method, and register instances.
    # Consider not storing an actual OperationImplementation in the registry.
    # Note, though, that if we want to automatically register on import via
    # base class (meta-class), the functionality must be in the class definition.
    @classmethod
    @abstractmethod
    def director(cls, context: Context) -> OperationDirector:
        """Factory to get an OperationDirector appropriate for the context."""
        ...



###########################################################################
#                                                                         #
#    physical_validation,                                                 #
#    a python package to test the physical validity of MD results         #
#                                                                         #
#    Written by Michael R. Shirts <michael.shirts@colorado.edu>           #
#               Pascal T. Merz <pascal.merz@colorado.edu>                 #
#                                                                         #
#    Copyright (C) 2012 University of Virginia                            #
#              (C) 2017 University of Colorado Boulder                    #
#                                                                         #
#    This library is free software; you can redistribute it and/or        #
#    modify it under the terms of the GNU Lesser General Public           #
#    License as published by the Free Software Foundation; either         #
#    version 2.1 of the License, or (at your option) any later version.   #
#                                                                         #
#    This library is distributed in the hope that it will be useful,      #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of       #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU    #
#    Lesser General Public License for more details.                      #
#                                                                         #
#    You should have received a copy of the GNU Lesser General Public     #
#    License along with this library; if not, write to the                #
#    Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,     #
#    Boston, MA 02110-1301 USA                                            #
#                                                                         #
###########################################################################
r"""
Data structures carrying simulation data.
"""
import warnings

from ..util import error as pv_error


class EnsembleData(object):
    r"""EnsembleData: Holds data defining the ensemble

    The ensemble is a string indicating the thermodynamical ensemble a simulation was
    performed in, and is any of 'NVE', 'NVT', 'NPT', 'muVT'.
    Depending on the ensemble, EnsembleData then holds additional information defining
    the ensemble, such as the number of particles N, the chemical potential mu, the
    volume V, the pressure P, the constant energy E or the temperature T. While any
    of these additional information are optional, most of them are needed by certain
    tests, such that not fully defining the ensemble results in warnings. The notable
    exception to this rule is the constant energy E for the NVE, which is not needed
    by any test and can hence be omitted without raising a warning.
    """

    @staticmethod
    def ensembles():
        return ('NVE',
                'NVT',
                'NPT',
                'muVT')

    def __init__(self, ensemble,
                 natoms=None, mu=None,
                 volume=None, pressure=None,
                 energy=None, temperature=None):
        self.__ensemble = None
        self.__n = None
        self.__mu = None
        self.__v = None
        self.__p = None
        self.__e = None
        self.__t = None

        if ensemble not in self.ensembles():
            raise pv_error.InputError('ensemble',
                                      'Given ensemble unknown.')
        self.__ensemble = ensemble

        if ensemble == 'NVE':
            if natoms is None:
                warnings.warn(ensemble + ' with undefined natoms.')
            if volume is None:
                warnings.warn(ensemble + ' with undefined volume.')
            # if energy is None:
            #     warnings.warn(ensemble + ' with undefined energy.')
            self.__n = natoms
            self.__v = volume
            self.__e = energy
        if ensemble == 'NVT':
            if natoms is None:
                warnings.warn(ensemble + ' with undefined natoms.')
            if volume is None:
                warnings.warn(ensemble + ' with undefined volume.')
            if temperature is None:
                warnings.warn(ensemble + ' with undefined temperature.')
            self.__n = natoms
            self.__v = volume
            self.__t = temperature
        if ensemble == 'NPT':
            if natoms is None:
                warnings.warn(ensemble + ' with undefined natoms.')
            if pressure is None:
                warnings.warn(ensemble + ' with undefined pressure.')
            if temperature is None:
                warnings.warn(ensemble + ' with undefined temperature.')
            self.__n = natoms
            self.__p = pressure
            self.__t = temperature
        if ensemble == 'muVT':
            if mu is None:
                warnings.warn(ensemble + ' with undefined mu.')
            if volume is None:
                warnings.warn(ensemble + ' with undefined volume.')
            if temperature is None:
                warnings.warn(ensemble + ' with undefined temperature.')
            self.__mu = mu
            self.__v = volume
            self.__t = temperature

    @property
    def ensemble(self):
        """Get ensemble"""
        return self.__ensemble

    @property
    def natoms(self):
        """Get natoms"""
        return self.__n

    @property
    def mu(self):
        """Get mu"""
        return self.__mu

    @property
    def volume(self):
        """Get volume"""
        return self.__v

    @property
    def pressure(self):
        """Get pressure"""
        return self.__p

    @property
    def energy(self):
        """Get energy"""
        return self.__e

    @property
    def temperature(self):
        """Get temperature"""
        return self.__t



#
# This file is part of the GROMACS molecular simulation package.
#
# Copyright 2019- The GROMACS Authors
# and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
# Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
#
# GROMACS is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# GROMACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with GROMACS; if not, see
# https://www.gnu.org/licenses, or write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
#
# If you want to redistribute modifications to GROMACS, please
# consider that scientific software is very special. Version
# control is crucial - bugs must be traceable. We will be happy to
# consider code for inclusion in the official distribution, but
# derived work must not be called official GROMACS. Details are found
# in the README & COPYING files - if they are missing, get the
# official version at https://www.gromacs.org.
#
# To help us fund GROMACS development, we humbly ask that you cite
# the research papers on the package. Check out https://www.gromacs.org.

"""Test the utilities in the gmxapi.runtime module."""

import logging

import gmxapi.utility
from gmxapi.runtime import (
    scoped_resources,
    BaseContext,
    ResourceRequirements,
    ResourceAssignment,
)

try:
    from mpi4py import MPI

    rank_number = MPI.COMM_WORLD.Get_rank()
    comm_size = MPI.COMM_WORLD.Get_size()
except ImportError:
    rank_number = 0
    comm_size = 1
    rank_tag = ""
    MPI = None
else:
    rank_tag = "rank{}:".format(rank_number)

logger = logging.getLogger()


def test_no_requested_comm():
    """Indicate multiple subtasks, but don't request any comm resources."""
    base = BaseContext.instance()
    assert base.communicator().Get_rank() == rank_number
    assert base.communicator().Get_size() == comm_size
    assignment = None
    # Indicate multiple subtasks, but don't request any comm resources.
    for num_tasks in range(1, comm_size + 1):
        requirements = ResourceRequirements(
            communication=tuple({} for _ in range(num_tasks))
        )
        with scoped_resources(base, requirements=requirements) as assignment:
            assert assignment._base_comm_size == base.communicator().Get_size()
            assert assignment.base_rank() == base.communicator().Get_rank()
            assert set(assignment._group_ids) == set(range(num_tasks))
            group_id = assignment._group_ids[rank_number]
            assert group_id >= 0
            assert group_id < num_tasks
            roots = assignment._roots
            assert len(set(roots) - {None}) == num_tasks
            assert set(roots) - {None} == set(range(num_tasks))
            if rank_number == 0:
                assert group_id == 0
            if rank_number == comm_size - 1:
                assert group_id == num_tasks - 1
            if assignment.is_subtask_root():
                _count = 1
            else:
                _count = 0
            num_roots = base.communicator().allreduce(_count)
            assert num_roots == num_tasks
            assert not assignment.subtask_id()
            assert not assignment.subtask_rank()
            assert set(assignment._group_ids) == set(range(num_tasks))
    assert isinstance(assignment, ResourceAssignment)
    message = " ".join(
        [
            f"*{rank}*" if rank == assignment.base_rank() else f"{rank}"
            for rank in range(comm_size)
        ]
    )
    logger.info(message)


def test_no_ensemble_comm():
    """Indicate multiple subtasks, but don't request an ensemble comm."""
    base = BaseContext.instance()
    for num_tasks in range(1, comm_size + 1):
        requirements = ResourceRequirements(
            communication=tuple({"subtask_comm": True} for _ in range(num_tasks))
        )
        with scoped_resources(base, requirements=requirements) as assignment:
            if assignment.subtask_rank() == 0:
                assert assignment.is_subtask_root()
            subtasks = base.communicator().allgather(assignment.subtask_id())
            subtasks = set(subtasks) - {None}
            assert subtasks == set(range(num_tasks))
            subtask_communicator = assignment.communicator()
            if assignment.subtask_rank() is not None:
                assert subtask_communicator.Get_rank() == assignment.subtask_rank()
                _count = 1
            else:
                _count = 0
            total = base.communicator().allreduce(_count)
            if gmxapi.utility.config()["gmx_mpi_type"] == "library":
                assert total == comm_size
            else:
                if subtask_communicator:
                    assert subtask_communicator.Get_size() == 1
                assert total == num_tasks
            if assignment.subtask_id():
                total = subtask_communicator.allreduce(_count)
                if gmxapi.utility.config()["gmx_mpi_type"] == "library":
                    assert total >= 1
                    assert total <= comm_size - (num_tasks - 1)
                else:
                    assert total == 1
                    assert assignment.subtask_rank() == 0


def test_no_subtask_comm():
    """Indicate multiple subtasks, but don't request subtask comms."""
    base = BaseContext.instance()
    for num_tasks in range(1, comm_size + 1):
        requirements = ResourceRequirements(
            communication=tuple({"ensemble_comm": 1} for _ in range(num_tasks))
        )
        with scoped_resources(base, requirements=requirements) as assignment:
            assert not assignment.subtask_id()
            assert not assignment.subtask_rank()
            assert assignment._resource_id is not None
            labels = base.communicator().allgather(assignment._resource_id)
            assert len(set(labels)) == 1
            if assignment.is_subtask_root():
                assert assignment.ensemble_communicator()
            if assignment.ensemble_communicator():
                assert assignment.ensemble_communicator().Get_size() == num_tasks
                assert assignment.is_subtask_root()
                assert (
                    assignment._group_ids[rank_number]
                    == assignment.ensemble_communicator().Get_rank()
                )
                _count = 1
                assert assignment.ensemble_communicator().allreduce(_count) == num_tasks
            else:
                _count = 0
            total = base.communicator().allreduce(_count)
            assert total == num_tasks


def test_assign_ensemble():
    """Indicate an ensemble of subtasks with all resources."""
    base = BaseContext.instance()
    # for num_tasks in range(3, 4):
    for num_tasks in range(1, comm_size + 1):
        requirements = ResourceRequirements(
            communication=tuple(
                {"ensemble_comm": 2, "subtask_comm": True} for _ in range(num_tasks)
            )
        )
        with scoped_resources(base, requirements=requirements) as assignment:
            subtask_rank = assignment.subtask_rank()
            if subtask_rank is not None:
                subtask_comm = assignment.communicator()
                assert subtask_comm.Get_rank() == subtask_rank
                assert subtask_comm.Get_size() > subtask_rank
            if subtask_rank == 0:
                assert bool(assignment.ensemble_communicator())
            if assignment.ensemble_communicator():
                assert subtask_rank == 0
                assert (
                    assignment.subtask_id()
                    == assignment.ensemble_communicator().Get_rank()
                )



#
# This file is part of the GROMACS molecular simulation package.
#
# Copyright 2019- The GROMACS Authors
# and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
# Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
#
# GROMACS is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# GROMACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with GROMACS; if not, see
# https://www.gnu.org/licenses, or write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
#
# If you want to redistribute modifications to GROMACS, please
# consider that scientific software is very special. Version
# control is crucial - bugs must be traceable. We will be happy to
# consider code for inclusion in the official distribution, but
# derived work must not be called official GROMACS. Details are found
# in the README & COPYING files - if they are missing, get the
# official version at https://www.gromacs.org.
#
# To help us fund GROMACS development, we humbly ask that you cite
# the research papers on the package. Check out https://www.gromacs.org.

"""Test gmx.fileio submodule"""
import os
import tempfile

import gmxapi
import pytest
from gmxapi.simulation.fileio import TprFile
from gmxapi.simulation.fileio import read_tpr
from gmxapi.exceptions import UsageError


@pytest.mark.usefixtures("cleandir")
def test_tprfile_read_old(spc_water_box):
    tpr_filename = spc_water_box
    with pytest.raises(UsageError):
        TprFile(tpr_filename, "x")
    with pytest.raises(UsageError):
        TprFile()
    tprfile = TprFile(tpr_filename, "r")
    with tprfile as fh:
        cpp_object = fh._tprFileHandle
        assert cpp_object is not None
        params = cpp_object.params().extract()
        assert "nsteps" in params
        assert "foo" not in params


@pytest.mark.usefixtures("cleandir")
def test_read_tpr(spc_water_box):
    tpr_filename = spc_water_box
    tpr_source = gmxapi.read_tpr(tpr_filename)
    assert "parameters" in tpr_source.output
    assert hasattr(tpr_source.output, "parameters")
    parameters = tpr_source.output.parameters.result()
    assert "nsteps" in parameters
    assert "foo" not in parameters
    assert parameters["nsteps"] == 2
    assert tpr_source.output.parameters["nsteps"].result() == 2


@pytest.mark.usefixtures("cleandir")
def test_write_tpr_file(spc_water_box):
    """Test gmx.fileio.write_tpr_file() using gmx.core API."""
    tpr_filename = spc_water_box
    additional_steps = 5000
    sim_input = read_tpr(tpr_filename)
    params = sim_input.parameters.extract()
    nsteps = params["nsteps"]
    init_step = params["init-step"]
    # Choose a new nsteps to check integer parameter setting.
    new_nsteps = init_step + additional_steps
    # Choose a new dt to check floating point parameter setting
    new_dt = params["dt"] * 2.0

    sim_input.parameters.set("nsteps", new_nsteps)
    sim_input.parameters.set("dt", new_dt)

    _, temp_filename = tempfile.mkstemp(suffix=".tpr")
    gmxapi.simulation.fileio.write_tpr_file(temp_filename, input=sim_input)
    tprfile = TprFile(temp_filename, "r")
    with tprfile as fh:
        params = read_tpr(fh).parameters.extract()
        # Note that we have chosen an exactly representable dt for spc_water_box.
        # Otherwise, we would have to use pytest.approx with a suitable tolerance.
        assert params["dt"] == new_dt
        assert params["nsteps"] != nsteps
        assert params["nsteps"] == new_nsteps

    os.unlink(temp_filename)



#
# This file is part of the GROMACS molecular simulation package.
#
# Copyright 2019- The GROMACS Authors
# and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
# Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
#
# GROMACS is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# GROMACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with GROMACS; if not, see
# https://www.gnu.org/licenses, or write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
#
# If you want to redistribute modifications to GROMACS, please
# consider that scientific software is very special. Version
# control is crucial - bugs must be traceable. We will be happy to
# consider code for inclusion in the official distribution, but
# derived work must not be called official GROMACS. Details are found
# in the README & COPYING files - if they are missing, get the
# official version at https://www.gromacs.org.
#
# To help us fund GROMACS development, we humbly ask that you cite
# the research papers on the package. Check out https://www.gromacs.org.

"""gmxapi Python package for GROMACS.

This package provides Python access to GROMACS molecular simulation tools.
Operations can be connected flexibly to allow high performance simulation and
analysis with complex control and data flows. Users can define new operations
in C++ or Python with the same tool kit used to implement this package.

"""

__all__ = [
    "commandline_operation",
    "concatenate_lists",
    "function_wrapper",
    "join_arrays",
    "logger",
    "logical_not",
    "make_constant",
    "mdrun",
    "modify_input",
    "ndarray",
    "read_tpr",
    "subgraph",
    "while_loop",
    "NDArray",
    "__version__",
]

from ._logging import logger
from .version import __version__

# Import utilities
from .operation import computed_result, function_wrapper

# Import public types
from .datamodel import NDArray

# Import the public operations
from .datamodel import ndarray
from .operation import concatenate_lists, join_arrays, logical_not, make_constant
from .commandline import commandline_operation
from .simulation import mdrun, modify_input, read_tpr

# TODO: decide where this lives
from .operation import subgraph

# TODO: decide where this lives
from .operation import while_loop



#!/usr/bin/env python3
#
# This file is part of the GROMACS molecular simulation package.
#
# Copyright 2014- The GROMACS Authors
# and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
# Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
#
# GROMACS is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# GROMACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with GROMACS; if not, see
# https://www.gnu.org/licenses, or write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
#
# If you want to redistribute modifications to GROMACS, please
# consider that scientific software is very special. Version
# control is crucial - bugs must be traceable. We will be happy to
# consider code for inclusion in the official distribution, but
# derived work must not be called official GROMACS. Details are found
# in the README & COPYING files - if they are missing, get the
# official version at https://www.gromacs.org.
#
# To help us fund GROMACS development, we humbly ask that you cite
# the research papers on the package. Check out https://www.gromacs.org.

"""GROMACS-specific representation for source tree and documentation.

This module provides classes that construct a GROMACS-specific representation
of the source tree and associate the Doxygen XML output with it.  It constructs
an initial representation by walking the source tree in the file system, and
then associates information from the Doxygen XML output into this.
It also adds some additional knowledge from how the GROMACS source tree is
organized to construct a representation that is easy to process and check as
the top-level scripts expect.

The object model is rooted at a GromacsTree object.  Currently, it constructs a
representation of the source tree from the file system, but is otherwise mostly
a thin wrapper around the Doxygen XML tree.  It already adds some relations and
rules that come from GROMACS-specific knowledge.  In the future, more such
customizations will be added.
"""

import collections
import os
import os.path
import re
import subprocess

import doxygenxml as xml
import reporter

# We import DocType directly so that it is exposed from this module as well.
from doxygenxml import DocType


def _get_api_type_for_compound(grouplist):
    """Helper function to deduce API type from Doxygen group membership."""
    result = DocType.internal
    for group in grouplist:
        if isinstance(group, xml.Group):
            if group.get_name() == "group_publicapi":
                result = DocType.public
            elif group.get_name() == "group_libraryapi":
                result = DocType.library
            # TODO: Check for multiple group membership
    return result


class IncludedFile(object):
    """Information about an #include directive in a file."""

    def __init__(
        self,
        including_file,
        lineno,
        included_file,
        included_path,
        is_relative,
        is_system,
        line,
    ):
        self._including_file = including_file
        self._line_number = lineno
        self._included_file = included_file
        self._included_path = included_path
        # self._used_include_path = used_include_path
        self._is_relative = is_relative
        self._is_system = is_system
        self._line = line

    def __str__(self):
        if self._is_system:
            return "<{0}>".format(self._included_path)
        else:
            return '"{0}"'.format(self._included_path)

    def __lt__(self, other):
        return str(self) < str(other)

    def is_system(self):
        return self._is_system

    def is_relative(self):
        return self._is_relative

    def get_included_path(self):
        return self._included_path

    def get_including_file(self):
        return self._including_file

    def get_file(self):
        return self._included_file

    def get_line_number(self):
        return self._line_number

    def get_full_line(self):
        """Return the full source line on which this include appears.

        Trailing newline is included."""
        return self._line

    def get_reporter_location(self):
        return reporter.Location(self._including_file.get_abspath(), self._line_number)


class IncludeBlock(object):
    """Block of consequent #include directives in a file."""

    def __init__(self, first_included_file):
        self._first_line = first_included_file.get_line_number()
        self._last_line = self._first_line
        self._files = []
        self.add_file(first_included_file)

    def add_file(self, included_file):
        self._files.append(included_file)
        self._last_line = included_file.get_line_number()

    def get_includes(self):
        return self._files

    def get_first_line(self):
        return self._first_line

    def get_last_line(self):
        return self._last_line


class File(object):
    """Source/header file in the GROMACS tree."""

    def __init__(self, abspath, relpath, directory):
        """Initialize a file representation with basic information."""
        self._abspath = abspath
        self._relpath = relpath
        self._dir = directory
        self._rawdoc = None
        extension = os.path.splitext(abspath)[1]
        self._sourcefile = extension in (".c", ".cc", ".cpp", ".cu")
        self._apitype = DocType.none
        self._modules = set()
        self._includes = []
        self._include_blocks = []
        self._main_header = None
        self._lines = None
        self._filter = None
        self._declared_defines = None
        self._used_defines = dict()
        directory.add_file(self)

    def set_doc_xml(self, rawdoc, sourcetree):
        """Assiociate Doxygen documentation entity with the file."""
        assert self._rawdoc is None
        assert rawdoc.is_source_file() == self._sourcefile
        self._rawdoc = rawdoc
        if self._rawdoc.is_documented():
            grouplist = self._rawdoc.get_groups()
            self._apitype = _get_api_type_for_compound(grouplist)
            for group in grouplist:
                module = sourcetree.get_object(group)
                if module:
                    self._modules.add(module)

    def set_git_filter_attribute(self, filtername):
        """Set the git filter attribute associated with the file."""
        self._filter = filtername

    def set_main_header(self, included_file):
        """Set the main header file for a source file."""
        assert self.is_source_file()
        self._main_header = included_file

    def _process_include(self, lineno, is_system, includedpath, line, sourcetree):
        """Process #include directive during scan()."""
        is_relative = False
        if is_system:
            fileobj = sourcetree.find_include_file(includedpath)
        else:
            fullpath = os.path.join(self._dir.get_abspath(), includedpath)
            fullpath = os.path.abspath(fullpath)
            if os.path.exists(fullpath):
                is_relative = True
                fileobj = sourcetree.get_file(fullpath)
            else:
                fileobj = sourcetree.find_include_file(includedpath)
        included_file = IncludedFile(
            self, lineno, fileobj, includedpath, is_relative, is_system, line
        )
        self._includes.append(included_file)
        return included_file

    def scan_contents(self, sourcetree, keep_contents, detect_defines):
        """Scan the file contents and initialize information based on it."""
        # TODO: Consider a more robust regex.
        include_re = r'^\s*#\s*include\s+(?P<quote>["<])(?P<path>[^">]*)[">]'
        define_re = r"^\s*#.*define(?:01)?\s+(\w*)"
        current_block = None
        with open(self._abspath, "r", encoding="utf8") as scanfile:
            contents = scanfile.read()
        lines = contents.splitlines(True)
        for lineno, line in enumerate(lines, 1):
            match = re.match(include_re, line)
            if match:
                is_system = match.group("quote") == "<"
                includedpath = match.group("path")
                included_file = self._process_include(
                    lineno, is_system, includedpath, line, sourcetree
                )
                if current_block is None:
                    current_block = IncludeBlock(included_file)
                    self._include_blocks.append(current_block)
                else:
                    current_block.add_file(included_file)
            elif line and not line.isspace():
                current_block = None
        if detect_defines:
            self._declared_defines = []
            for line in lines:
                match = re.match(define_re, line)
                if match:
                    self._declared_defines.append(match.group(1))
        if keep_contents:
            self._lines = lines

    def add_used_defines(self, define_file, defines):
        """Add defines used in this file.

        Used internally by find_define_file_uses()."""
        if define_file not in self._used_defines:
            self._used_defines[define_file] = set()
        self._used_defines[define_file].update(defines)

    def get_reporter_location(self):
        return reporter.Location(self._abspath, None)

    def is_external(self):
        return self._dir.is_external()

    def is_source_file(self):
        return self._sourcefile

    def is_test_file(self):
        return self._dir.is_test_directory()

    def should_includes_be_sorted(self):
        """Return whether the include directives in the file should be sorted."""
        return self._filter in ("includesort", "complete_formatting")

    def is_documented(self):
        return self._rawdoc and self._rawdoc.is_documented()

    def has_brief_description(self):
        return self._rawdoc and self._rawdoc.has_brief_description()

    def get_abspath(self):
        return self._abspath

    def get_relpath(self):
        return self._relpath

    def get_name(self):
        return os.path.basename(self._abspath)

    def get_directory(self):
        return self._dir

    def get_doc_type(self):
        if not self._rawdoc:
            return DocType.none
        return self._rawdoc.get_visibility()

    def get_api_type(self):
        return self._apitype

    def api_type_is_reliable(self):
        if self._apitype in (DocType.internal, DocType.library):
            return True
        module = self.get_module()
        return module and module.is_documented()

    def is_public(self):
        return self.get_api_type() == DocType.public

    def is_module_internal(self):
        if self.is_source_file():
            return True
        return self.get_api_type() <= DocType.internal

    def get_expected_module(self):
        return self._dir.get_module()

    def get_doc_modules(self):
        return self._modules

    def get_module(self):
        module = self.get_expected_module()
        if not module and len(self._modules) == 1:
            module = list(self._modules)[0]
        return module

    def get_includes(self):
        return self._includes

    def get_include_blocks(self):
        return self._include_blocks

    def _get_included_files_recurse(self, result):
        for include in self._includes:
            included_file = include.get_file()
            if included_file is not None and not included_file in result:
                result.add(included_file)
                included_file._get_included_files_recurse(result)

    def get_included_files(self, recursive=False):
        if recursive:
            result = set()
            self._get_included_files_recurse(result)
            return result
        return set([x.get_file() for x in self._includes])

    def get_main_header(self):
        return self._main_header

    def get_contents(self):
        return self._lines

    def get_declared_defines(self):
        """Return set of defines declared in this file.

        The information is only populated for selected files."""
        return self._declared_defines

    def get_used_define_files(self):
        """Return files like config.h whose defines are used in this file.

        The return value is empty if find_define_file_uses() has not been called,
        as well as for headers that declare these defines."""
        return set(self._used_defines.keys())

    def get_used_defines(self, define_file):
        """Return set of defines used in this file for a given file like config.h."""
        return self._used_defines.get(define_file, set())


class GeneratedFile(File):
    def __init__(self, abspath, relpath, directory):
        File.__init__(self, abspath, relpath, directory)
        self._generator_source_file = None

    def scan_contents(self, sourcetree, keep_contents, detect_defines):
        if os.path.exists(self.get_abspath()):
            File.scan_contents(self, sourcetree, keep_contents, False)

    def set_generator_source(self, sourcefile):
        self._generator_source_file = sourcefile

    def get_generator_source(self):
        return self._generator_source_file

    def get_reporter_location(self):
        if self._generator_source_file:
            return self._generator_source_file.get_reporter_location()
        return File.get_reporter_location(self)

    def get_declared_defines(self):
        if self._generator_source_file:
            return self._generator_source_file.get_declared_defines()
        return File.get_declared_defines(self)


class GeneratorSourceFile(File):
    pass


class Directory(object):
    """(Sub)directory in the GROMACS tree."""

    def __init__(self, abspath, relpath, parent):
        """Initialize a file representation with basic information."""
        self._abspath = abspath
        self._relpath = relpath
        self._name = os.path.basename(abspath)
        self._parent = parent
        self._rawdoc = None
        self._module = None
        self._is_test_dir = False
        if parent and parent.is_test_directory() or self._name == "tests":
            self._is_test_dir = True
        self._is_external = False
        if parent and parent.is_external() or self._name == "external":
            self._is_external = True
        self._subdirs = set()
        if parent:
            parent._subdirs.add(self)
        self._files = set()

    def set_doc_xml(self, rawdoc, sourcetree):
        """Assiociate Doxygen documentation entity with the directory."""
        assert self._rawdoc is None
        assert rawdoc.get_path().rstrip("/") in (self._abspath, self._relpath)
        self._rawdoc = rawdoc

    def set_module(self, module):
        assert self._module is None
        self._module = module

    def add_file(self, fileobj):
        self._files.add(fileobj)

    def get_name(self):
        return self._name

    def get_reporter_location(self):
        return reporter.Location(self._abspath, None)

    def get_abspath(self):
        return self._abspath

    def get_relpath(self):
        return self._relpath

    def is_test_directory(self):
        return self._is_test_dir

    def is_external(self):
        return self._is_external

    def get_module(self):
        if self._module:
            return self._module
        if self._parent:
            return self._parent.get_module()
        return None

    def get_subdirectories(self):
        return self._subdirs

    def get_files(self):
        for subdir in self._subdirs:
            for fileobj in subdir.get_files():
                yield fileobj
        for fileobj in self._files:
            yield fileobj

    def contains(self, fileobj):
        """Check whether file is within the directory or its subdirectories."""
        dirobj = fileobj.get_directory()
        while dirobj:
            if dirobj == self:
                return True
            dirobj = dirobj._parent
        return False


class ModuleDependency(object):
    """Dependency between modules."""

    def __init__(self, othermodule):
        """Initialize empty dependency object with given module as dependency."""
        self._othermodule = othermodule
        self._includedfiles = []
        self._cyclesuppression = None
        self._is_test_only_dependency = True
        self.suppression_used = True

    def add_included_file(self, includedfile):
        """Add IncludedFile that is part of this dependency."""
        assert includedfile.get_file().get_module() == self._othermodule
        if not includedfile.get_including_file().is_test_file():
            self._is_test_only_dependency = False
        self._includedfiles.append(includedfile)

    def set_cycle_suppression(self):
        """Set suppression on cycles containing this dependency."""
        self._cyclesuppression = True
        self.suppression_used = False

    def is_cycle_suppressed(self):
        """Return whether cycles containing this dependency are suppressed."""
        self.suppression_used = True
        return self._cyclesuppression is not None

    def is_test_only_dependency(self):
        """Return whether this dependency is only from test code."""
        return self._is_test_only_dependency

    def get_other_module(self):
        """Get module that this dependency is to."""
        return self._othermodule

    def get_included_files(self):
        """Get IncludedFile objects for the individual include dependencies."""
        return self._includedfiles


class Module(object):
    """Code module in the GROMACS source tree.

    Modules are specific subdirectories that host a more or less coherent
    set of routines.  Simplified, every subdirectory under src/gromacs/ is
    a different module.  This object provides that abstraction and also links
    the subdirectory to the module documentation (documented as a group in
    Doxygen) if that exists.
    """

    def __init__(self, name, rootdir):
        self._name = name
        self._rawdoc = None
        self._rootdir = rootdir
        self._group = None
        self._dependencies = dict()

    def set_doc_xml(self, rawdoc, sourcetree):
        """Assiociate Doxygen documentation entity with the module."""
        assert self._rawdoc is None
        self._rawdoc = rawdoc
        if self._rawdoc.is_documented():
            groups = list(self._rawdoc.get_groups())
            if len(groups) == 1:
                groupname = groups[0].get_name()
                if groupname.startswith("group_"):
                    self._group = groupname[6:]

    def add_dependency(self, othermodule, includedfile):
        """Add #include dependency from a file in this module."""
        assert includedfile.get_file().get_module() == othermodule
        if othermodule not in self._dependencies:
            self._dependencies[othermodule] = ModuleDependency(othermodule)
        self._dependencies[othermodule].add_included_file(includedfile)

    def is_documented(self):
        return self._rawdoc is not None

    def get_name(self):
        return self._name

    def get_root_dir(self):
        return self._rootdir

    def get_files(self):
        # TODO: Include public API convenience headers?
        return self._rootdir.get_files()

    def get_group(self):
        return self._group

    def get_dependencies(self):
        return self._dependencies.values()


class Namespace(object):
    """Namespace in the GROMACS source code."""

    def __init__(self, rawdoc):
        self._rawdoc = rawdoc

    def is_anonymous(self):
        return self._rawdoc.is_anonymous()


class Class(object):
    """Class/struct/union in the GROMACS source code."""

    def __init__(self, rawdoc, files):
        self._rawdoc = rawdoc
        self._files = set(files)

    def get_name(self):
        return self._rawdoc.get_name()

    def get_reporter_location(self):
        return self._rawdoc.get_reporter_location()

    def get_files(self):
        return self._files

    def is_documented(self):
        return self._rawdoc.is_documented()

    def has_brief_description(self):
        return self._rawdoc.has_brief_description()

    def get_doc_type(self):
        """Return documentation type (visibility) for the class.

        In addition to the actual code, this encodes GROMACS-specific logic
        of setting EXTRACT_LOCAL_CLASSES=YES only for the full documentation.
        Local classes never appear outside the full documentation, no matter
        what is their visibility.
        """
        if not self.is_documented():
            return DocType.none
        if self._rawdoc.is_local():
            return DocType.internal
        return self._rawdoc.get_visibility()

    def get_file_doc_type(self):
        return max([fileobj.get_doc_type() for fileobj in self._files])


class Member(object):
    """Member (in Doxygen terminology) in the GROMACS source tree.

    Currently, modeling is limited to the minimal set of properties that the
    checker uses.
    """

    def __init__(self, rawdoc, namespace):
        self._rawdoc = rawdoc
        self._namespace = namespace

    def get_name(self):
        return self._rawdoc.get_name()

    def get_reporter_location(self):
        return self._rawdoc.get_reporter_location()

    def is_documented(self):
        return self._rawdoc.is_documented()

    def has_brief_description(self):
        return self._rawdoc.has_brief_description()

    def has_inbody_description(self):
        return self._rawdoc.has_inbody_description()

    def is_visible(self):
        """Return whether the member is visible in Doxygen documentation.

        Doxygen ignores members whose parent compounds are not documented.
        However, when EXTRACT_ANON_NPACES=ON (which is set for our full
        documentation), members of anonymous namespaces are extracted even if
        the namespace is the only parent and is not documented.
        """
        if self._namespace and self._namespace.is_anonymous():
            return True
        return self._rawdoc.get_inherited_visibility() != DocType.none


class GromacsTree(object):
    """Root object for navigating the GROMACS source tree.

    On initialization, the list of files and directories is initialized by
    walking the source tree, and modules are created for top-level
    subdirectories.  At this point, only information that is accessible from
    file names and paths only is available.

    load_git_attributes() can be called to load attribute information from
    .gitattributes for all the files.

    scan_files() can be called to read all the files and initialize #include
    dependencies between the files based on the information.  This is done like
    this instead of relying on Doxygen-extracted include files to make the
    dependency graph independent from preprocessor macro definitions
    (Doxygen only sees those #includes that the preprocessor sees, which
    depends on what #defines it has seen).

    find_define_file_uses() can be called to find all uses of defines
    declared in config.h and some other macro headers. In the current
    implementation, scan_files() must have been called earlier.

    load_xml() can be called to load information from Doxygen XML data in
    the build tree (the Doxygen XML data must have been built separately).
    """

    def __init__(self, source_root, build_root, reporter):
        """Initialize the tree object by walking the source tree."""
        self._source_root = os.path.abspath(source_root)
        self._build_root = os.path.abspath(build_root)
        self._reporter = reporter
        self._docset = None
        self._docmap = dict()
        self._dirs = dict()
        self._files = dict()
        self._modules = dict()
        self._classes = set()
        self._namespaces = set()
        self._members = set()
        self._walk_dir(os.path.join(self._source_root, "src"))
        for fileobj in self.get_files():
            if fileobj and fileobj.is_source_file() and not fileobj.is_external():
                (basedir, name) = os.path.split(fileobj.get_abspath())
                (basename, ext) = os.path.splitext(name)
                header = self.get_file(os.path.join(basedir, basename + ".h"))
                if not header and ext == ".cu":
                    header = self.get_file(os.path.join(basedir, basename + ".cuh"))
                if not header and fileobj.is_test_file():
                    basedir = os.path.dirname(basedir)
                    header = self.get_file(os.path.join(basedir, basename + ".h"))
                    if not header:
                        # Somewhat of a hack; currently, the tests for
                        # analysisdata/modules/ and trajectoryanalysis/modules/
                        # is at the top-level tests directory.
                        # TODO: It could be clearer to split the tests so that
                        # there would be a separate modules/tests/.
                        header = self.get_file(
                            os.path.join(basedir, "modules", basename + ".h")
                        )
                    if not header and basename.endswith("_tests"):
                        header = self.get_file(
                            os.path.join(basedir, basename[:-6] + ".h")
                        )
                if header:
                    fileobj.set_main_header(header)
        rootdir = self._get_dir(os.path.join("src", "gromacs"))
        for subdir in rootdir.get_subdirectories():
            self._create_module(subdir)
        # list of directories that aren't in the default location src/gromacs
        rootdir = self._get_dir(os.path.join("src", "testutils"))
        self._create_module(rootdir)
        rootdir = self._get_dir(os.path.join("src", "gromacs", "applied_forces", "awh"))
        self._create_module(rootdir)
        rootdir = self._get_dir(os.path.join("src", "programs", "mdrun", "tests"))
        self._create_module(rootdir)

    def _get_rel_path(self, path):
        assert os.path.isabs(path)
        if path.startswith(self._build_root):
            return os.path.relpath(path, self._build_root)
        if path.startswith(self._source_root):
            return os.path.relpath(path, self._source_root)
        raise ValueError("path not under build nor source tree: {0}".format(path))

    def _walk_dir(self, rootpath):
        """Construct representation of the source tree by walking the file system."""
        assert os.path.isabs(rootpath)
        assert rootpath not in self._dirs
        relpath = self._get_rel_path(rootpath)
        self._dirs[relpath] = Directory(rootpath, relpath, None)
        for dirpath, dirnames, filenames in os.walk(rootpath):
            if "refdata" in dirnames:
                dirnames.remove("refdata")
            currentdir = self._dirs[self._get_rel_path(dirpath)]
            # Loop through a copy so that we can modify dirnames.
            for dirname in list(dirnames):
                fullpath = os.path.join(dirpath, dirname)
                if fullpath == self._build_root:
                    dirnames.remove(dirname)
                    continue
                relpath = self._get_rel_path(fullpath)
                self._dirs[relpath] = Directory(fullpath, relpath, currentdir)
            extensions = (".h", ".cuh", ".hpp", ".c", ".cc", ".cpp", ".cu", ".bm")
            for filename in filenames:
                basename, extension = os.path.splitext(filename)
                if extension in extensions:
                    fullpath = os.path.join(dirpath, filename)
                    relpath = self._get_rel_path(fullpath)
                    self._files[relpath] = File(fullpath, relpath, currentdir)
                elif extension == ".cmakein":
                    extension = os.path.splitext(basename)[1]
                    if extension in extensions:
                        fullpath = os.path.join(dirpath, filename)
                        relpath = self._get_rel_path(fullpath)
                        sourcefile = GeneratorSourceFile(fullpath, relpath, currentdir)
                        self._files[relpath] = sourcefile
                        fullpath = os.path.join(dirpath, basename)
                        relpath = self._get_rel_path(fullpath)
                        fullpath = os.path.join(self._build_root, relpath)
                        generatedfile = GeneratedFile(fullpath, relpath, currentdir)
                        self._files[relpath] = generatedfile
                        generatedfile.set_generator_source(sourcefile)
                elif extension in (".l", ".y", ".pre"):
                    fullpath = os.path.join(dirpath, filename)
                    relpath = self._get_rel_path(fullpath)
                    self._files[relpath] = GeneratorSourceFile(
                        fullpath, relpath, currentdir
                    )

    def _create_module(self, rootdir):
        """Create module for a subdirectory."""
        name = "module_" + rootdir.get_name()
        # Fix for module_mdrun_integration_tests
        # that contains files from three directories.
        # Here we use the directory that contains
        # nearly all files that belong to the module
        if name == "module_tests":
            name = "module_mdrun_integration_tests"
        moduleobj = Module(name, rootdir)
        rootdir.set_module(moduleobj)
        self._modules[name] = moduleobj

    def scan_files(self, only_files=None, keep_contents=False):
        """Read source files to initialize #include dependencies."""
        if only_files:
            filelist = only_files
        else:
            filelist = self._files.values()
        define_files = list(self.get_checked_define_files())
        for define_file in list(define_files):
            if (
                isinstance(define_file, GeneratedFile)
                and define_file.get_generator_source() is not None
            ):
                define_files.append(define_file.get_generator_source())
        for fileobj in filelist:
            if not fileobj.is_external():
                detect_defines = fileobj in define_files
                fileobj.scan_contents(self, keep_contents, detect_defines)
                module = fileobj.get_module()
                if module:
                    for includedfile in fileobj.get_includes():
                        otherfile = includedfile.get_file()
                        if otherfile:
                            othermodule = otherfile.get_module()
                            if othermodule and othermodule != module:
                                module.add_dependency(othermodule, includedfile)

    def load_xml(self, only_files=None):
        """Load Doxygen XML information.

        If only_files is True, XML data is not loaded for code constructs, but
        only for files, directories, and their potential parents.
        """
        xmldir = os.path.join(self._build_root, "docs", "html", "doxygen", "xml")
        self._docset = xml.DocumentationSet(xmldir, self._reporter)
        if only_files:
            if isinstance(only_files, collections.abc.Iterable):
                filelist = [x.get_relpath() for x in only_files]
                self._docset.load_file_details(filelist)
            else:
                self._docset.load_file_details()
        else:
            self._docset.load_details()
            self._docset.merge_duplicates()
        self._load_dirs()
        self._load_modules()
        self._load_files()
        if not only_files:
            self._load_namespaces()
            self._load_classes()
            self._load_members()

    def _load_dirs(self):
        """Load Doxygen XML directory information."""
        rootdirs = self._docset.get_compounds(
            xml.Directory, lambda x: x.get_parent() is None
        )
        for dirdoc in rootdirs:
            self._load_dir(dirdoc, None)

    def _load_dir(self, dirdoc, parent):
        """Load Doxygen XML directory information for a single directory."""
        path = dirdoc.get_path().rstrip("/")
        if not os.path.isabs(path):
            path = os.path.join(self._source_root, path)
        relpath = self._get_rel_path(path)
        dirobj = self._dirs.get(relpath)
        if not dirobj:
            dirobj = Directory(path, relpath, parent)
            self._dirs[relpath] = dirobj
        dirobj.set_doc_xml(dirdoc, self)
        self._docmap[dirdoc] = dirobj
        for subdirdoc in dirdoc.get_subdirectories():
            self._load_dir(subdirdoc, dirobj)

    def _load_modules(self):
        """Load Doxygen XML module (group) information."""
        moduledocs = self._docset.get_compounds(
            xml.Group, lambda x: x.get_name().startswith("module_")
        )
        for moduledoc in moduledocs:
            moduleobj = self._modules.get(moduledoc.get_name())
            if not moduleobj:
                self._reporter.input_error(
                    "no matching directory for module: {0}".format(moduledoc)
                )
                continue
            moduleobj.set_doc_xml(moduledoc, self)
            self._docmap[moduledoc] = moduleobj

    def _load_files(self):
        """Load Doxygen XML file information."""
        for filedoc in self._docset.get_files():
            path = filedoc.get_path()
            if not path:
                # In case of only partially loaded file information,
                # the path information is not set for unloaded files.
                continue
            if not os.path.isabs(path):
                path = os.path.join(self._source_root, path)
            extension = os.path.splitext(path)[1]
            # We don't care about Markdown files that only produce pages
            # (and fail the directory check below).
            if extension == ".md":
                continue
            dirdoc = filedoc.get_directory()
            if not dirdoc:
                self._reporter.xml_assert(
                    filedoc.get_xml_path(), "file is not in any directory in Doxygen"
                )
                continue
            relpath = self._get_rel_path(path)
            fileobj = self._files.get(relpath)
            if not fileobj:
                fileobj = File(path, relpath, self._docmap[dirdoc])
                self._files[relpath] = fileobj
            fileobj.set_doc_xml(filedoc, self)
            self._docmap[filedoc] = fileobj

    def _load_namespaces(self):
        """Load Doxygen XML namespace information."""
        nsdocs = self._docset.get_namespaces()
        for nsdoc in nsdocs:
            nsobj = Namespace(nsdoc)
            self._docmap[nsdoc] = nsobj
            self._namespaces.add(nsobj)

    def _load_classes(self):
        """Load Doxygen XML class information."""
        classdocs = self._docset.get_classes()
        for classdoc in classdocs:
            files = [self._docmap[filedoc] for filedoc in classdoc.get_files()]
            classobj = Class(classdoc, files)
            self._docmap[classdoc] = classobj
            self._classes.add(classobj)

    def _load_members(self):
        """Load Doxygen XML member information."""
        memberdocs = self._docset.get_members()
        for memberdoc in memberdocs:
            nsdoc = memberdoc.get_namespace()
            nsobj = self.get_object(nsdoc)
            memberobj = Member(memberdoc, nsobj)
            self._docmap[memberdoc] = memberobj
            self._members.add(memberobj)

    def _get_dir(self, relpath):
        """Get directory object for a path relative to source tree root."""
        return self._dirs.get(relpath)

    def get_file(self, path):
        """Get file object for a path relative to source tree root."""
        return self._files.get(self._get_rel_path(path))

    def find_include_file(self, includedpath):
        """Find a file object corresponding to an include path."""
        for testdir in (
            "src",
            "src/external/thread_mpi/include",
            "src/external/tng_io/include",
        ):
            testpath = os.path.join(testdir, includedpath)
            if testpath in self._files:
                return self._files[testpath]

    def load_git_attributes(self):
        """Load git attribute information for files."""
        args = ["git", "check-attr", "--stdin", "filter"]
        filelist = "\n".join(map(File.get_relpath, self._files.values())) + "\n"
        git_check_attr = subprocess.run(
            args,
            input=filelist,
            stdout=subprocess.PIPE,
            universal_newlines=True,
            cwd=self._source_root,
        )
        filters = git_check_attr.stdout
        for fileinfo in filters.splitlines():
            path, dummy, value = fileinfo.split(": ")
            fileobj = self._files.get(path)
            assert fileobj is not None
            fileobj.set_git_filter_attribute(value)

    def find_define_file_uses(self):
        """Find files that use defines from config.h."""
        # Executing git grep is substantially faster than using the define_re
        # directly on the contents of the file in Python.
        for define_file in self.get_checked_define_files():
            excluded_files = set([define_file])
            excluded_files.update(define_file.get_included_files(recursive=True))
            all_defines = define_file.get_declared_defines()
            args = ["git", "grep", "-zwIF"]
            for define in all_defines:
                args.extend(["-e", define])
            args.extend(["--", "*.cpp", "*.c", "*.cu", "*.h", "*.cuh"])
            define_re = r"\b(?:" + "|".join(all_defines) + r")\b"
            completed_process = subprocess.run(
                args,
                cwd=self._source_root,
                universal_newlines=True,
                stdout=subprocess.PIPE,
            )
            for line in completed_process.stdout.splitlines():
                (filename, text) = line.split("\0")
                fileobj = self._files.get(filename)
                if fileobj is not None and fileobj not in excluded_files:
                    defines = re.findall(define_re, text)
                    fileobj.add_used_defines(define_file, defines)

    def load_cycle_suppression_list(self, filename):
        """Load a list of edges to suppress in cycles.

        These edges between modules, if present, will be marked in the
        corresponding ModuleDependency objects.
        """
        with open(filename, "r") as fp:
            for line in fp:
                line = line.strip()
                if not line or line.startswith("#"):
                    continue
                modulenames = ["module_" + x.strip() for x in line.split("->")]
                if len(modulenames) != 2:
                    self._reporter.input_error(
                        "invalid cycle suppression line: {0}".format(line)
                    )
                    continue
                firstmodule = self._modules.get(modulenames[0])
                secondmodule = self._modules.get(modulenames[1])
                if not firstmodule or not secondmodule:
                    self._reporter.input_error(
                        "unknown modules mentioned on cycle suppression line: {0}".format(
                            line
                        )
                    )
                    continue
                for dep in firstmodule.get_dependencies():
                    if dep.get_other_module() == secondmodule:
                        dep.set_cycle_suppression()
                        break
                else:
                    self._reporter.cyclic_issue(
                        "unused cycle suppression: {0}".format(line)
                    )

    def report_unused_cycle_suppressions(self, reporter):
        """Reports unused cycle suppressions."""
        for module in self.get_modules():
            for dep in module.get_dependencies():
                if not dep.suppression_used:
                    reporter.cyclic_issue(
                        "unused cycle suppression: {0} -> {1}".format(
                            module.get_name()[7:], dep.get_other_module().get_name()[7:]
                        )
                    )

    def get_object(self, docobj):
        """Get tree object for a Doxygen XML object."""
        if docobj is None:
            return None
        return self._docmap.get(docobj)

    def get_files(self):
        """Get iterable for all files in the source tree."""
        return self._files.values()

    def get_modules(self):
        """Get iterable for all modules in the source tree."""
        return self._modules.values()

    def get_classes(self):
        """Get iterable for all classes in the source tree."""
        return self._classes

    def get_members(self):
        """Get iterable for all members (in Doxygen terms) in the source tree."""
        return self._members

    def get_checked_define_files(self):
        """Get list of files that contain #define macros whose usage needs to
        be checked."""
        return (
            self._files["src/config.h"],
            self._files["src/gromacs/simd/include/gromacs/simd/simd.h"],
            self._files["src/gromacs/ewald/pme_simd.h"],
            self._files["src/gromacs/nbnxm/nbnxm_simd.h"],
        )



###########################################################################
#                                                                         #
#    physical_validation,                                                 #
#    a python package to test the physical validity of MD results         #
#                                                                         #
#    Written by Michael R. Shirts <michael.shirts@colorado.edu>           #
#               Pascal T. Merz <pascal.merz@colorado.edu>                 #
#                                                                         #
#    Copyright (C) 2012 University of Virginia                            #
#              (C) 2017 University of Colorado Boulder                    #
#                                                                         #
#    This library is free software; you can redistribute it and/or        #
#    modify it under the terms of the GNU Lesser General Public           #
#    License as published by the Free Software Foundation; either         #
#    version 2.1 of the License, or (at your option) any later version.   #
#                                                                         #
#    This library is distributed in the hope that it will be useful,      #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of       #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU    #
#    Lesser General Public License for more details.                      #
#                                                                         #
#    You should have received a copy of the GNU Lesser General Public     #
#    License along with this library; if not, write to the                #
#    Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,     #
#    Boston, MA 02110-1301 USA                                            #
#                                                                         #
###########################################################################
r"""
flatfile_parser.py
"""
from . import parser
from . import SimulationData, TrajectoryData, ObservableData


class FlatfileParser(parser.Parser):
    """
    FlatfileParser
    """

    def __init__(self):
        super(FlatfileParser, self).__init__()

    def get_simulation_data(self, units=None, ensemble=None, system=None, dt=None,
                            position_file=None, velocity_file=None,
                            kinetic_ene_file=None, potential_ene_file=None,
                            total_ene_file=None, volume_file=None,
                            pressure_file=None, temperature_file=None,
                            const_of_mot_file=None):
        r"""Read simulation data from flat files

        Returns a SimulationData object created from (optionally) provided UnitData, EnsembleData
        and SystemData, as well as TrajectoryData and ObservableData objects created from flat
        files. The files are expected to be in one of the following formats:

        * xyz-format
          trajectory files (position_file, velocity_file)
          - three numbers per line, separated by white space
          - frames delimited by a completely blank line
          - any character after (and including) a '#' are ignored
        * 1d-format
          all other files
          - one number per line
          - any character after (and including) a '#' are ignored

        Parameters
        ----------
        units: UnitData, optional
            A UnitData object representing the units used in the simulation
        ensemble: EnsembleData, optional
            A EnsembleData object representing the ensemble the simulation has been performed in
        system: SystemData, optional
            A SystemData object representing the atoms and molecules in the system
        dt: float, optional
            The time step used in the simulation
        position_file: str, optional
            Path to a file in xyz-format containing the position trajectory
        velocity_file: str, optional
            Path to a file in xyz-format containing the velocity trajectory
        kinetic_ene_file: str, optional
            Path to a file in 1d-format containing the kinetic energy trajectory
        potential_ene_file: str, optional
            Path to a file in 1d-format containing the potential energy trajectory
        total_ene_file: str, optional
            Path to a file in 1d-format containing the total energy trajectory
        volume_file: str, optional
            Path to a file in 1d-format containing the volume trajectory
        pressure_file: str, optional
            Path to a file in 1d-format containing the pressure trajectory
        temperature_file: str, optional
            Path to a file in 1d-format containing the temperature trajectory
        const_of_mot_file: str, optional
            Path to a file in 1d-format containing the constant of motion trajectory

        Returns
        -------
        result: SimulationData
            A SimulationData filled with the provided ensemble and
            system objects as well as the trajectory data found in the
            edr and trr / gro files.

        """

        trj_dict = {
            'position': position_file,
            'velocity': velocity_file
        }

        if any(trj_dict.values()):
            trajectory = TrajectoryData()
            for key, filename in trj_dict.items():
                if filename is None:
                    continue
                trajectory[key] = self.__read_xyz(filename)
        else:
            trajectory = None

        obs_dict = {
            'kinetic_energy': kinetic_ene_file,
            'potential_energy': potential_ene_file,
            'total_energy': total_ene_file,
            'volume': volume_file,
            'pressure': pressure_file,
            'temperature': temperature_file,
            'constant_of_motion': const_of_mot_file
        }

        if any(obs_dict.values()):
            observables = ObservableData()
            for key, filename in obs_dict.items():
                if filename is None:
                    continue
                observables[key] = self.__read_1d(filename)
        else:
            observables = None

        result = SimulationData(units=units, dt=dt, system=system, ensemble=ensemble,
                                observables=observables, trajectory=trajectory)

        return result

    @staticmethod
    def __read_xyz(filename):
        result = []
        with open(filename) as f:
            frame = []
            for line in f:
                if not line:
                    # blank line
                    if frame:
                        result.append(frame)
                        frame = []
                    continue
                line = line.split('#', maxsplit=1)[0].strip()
                if not line:
                    # only comment on this line
                    continue
                xyz = line.split()[0:3]
                frame.append([float(n) for n in xyz])
            if frame:
                result.append(frame)
        return result

    @staticmethod
    def __read_1d(filename):
        result = []
        with open(filename) as f:
            for line in f:
                line = line.split('#', maxsplit=1)[0].strip()
                if not line:
                    # blank or comment-only line
                    continue
                result.append(float(line.strip()))
        return result



#
# This file is part of the GROMACS molecular simulation package.
#
# Copyright 2019- The GROMACS Authors
# and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
# Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
#
# GROMACS is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# GROMACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with GROMACS; if not, see
# https://www.gnu.org/licenses, or write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
#
# If you want to redistribute modifications to GROMACS, please
# consider that scientific software is very special. Version
# control is crucial - bugs must be traceable. We will be happy to
# consider code for inclusion in the official distribution, but
# derived work must not be called official GROMACS. Details are found
# in the README & COPYING files - if they are missing, get the
# official version at https://www.gromacs.org.
#
# To help us fund GROMACS development, we humbly ask that you cite
# the research papers on the package. Check out https://www.gromacs.org.

"""GROMACS simulation subpackage for gmxapi.

Provides operations for configuring and running molecular simulations.

The initial version of this module is a port of the gmxapi 0.0.7 facilities from
https://github.com/kassonlab/gmxapi and is not completely integrated with the
gmxapi 0.1 specification. Operation execution is dispatched to the old execution
manager for effective ensemble handling and C++ MD module binding. This should
be an implementation detail that is not apparent to the typical user, but it is
worth noting that chains of gmxapi.simulation module operations will be
automatically bundled for execution as gmxapi 0.0.7 style API sessions. Run time
options and file handling will necessarily change as gmxapi data flow handling
evolves.

In other words, if you rely on behavior not specified explicitly in the user
documentation, please keep an eye on the module documentation when updating
gmxapi and please participate in the ongoing discussions for design and
implementation.
"""

__all__ = ["abc", "mdrun", "modify_input", "read_tpr"]

from gmxapi.simulation import abc
from .mdrun import mdrun
from .read_tpr import read_tpr
from .modify_input import modify_input



###########################################################################
#                                                                         #
#    physical_validation,                                                 #
#    a python package to test the physical validity of MD results         #
#                                                                         #
#    Written by Michael R. Shirts <michael.shirts@colorado.edu>           #
#               Pascal T. Merz <pascal.merz@colorado.edu>                 #
#                                                                         #
#    Copyright (C) 2012 University of Virginia                            #
#              (C) 2017 University of Colorado Boulder                    #
#                                                                         #
#    This library is free software; you can redistribute it and/or        #
#    modify it under the terms of the GNU Lesser General Public           #
#    License as published by the Free Software Foundation; either         #
#    version 2.1 of the License, or (at your option) any later version.   #
#                                                                         #
#    This library is distributed in the hope that it will be useful,      #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of       #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU    #
#    Lesser General Public License for more details.                      #
#                                                                         #
#    You should have received a copy of the GNU Lesser General Public     #
#    License along with this library; if not, write to the                #
#    Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,     #
#    Boston, MA 02110-1301 USA                                            #
#                                                                         #
###########################################################################
r"""
Data structures carrying simulation data.
"""


class UnitData(object):
    r"""UnitData: Information about the units used

    The information about units consists of different parts:

    * The name of the units (energy_str, length_str, volume_str,
                             temperature_str, pressure_str, time_str),
    * the value of kB in the used energy units, and
    * the conversion factor to GROMACS units (kJ/mol, nm, nm^3, K, bar, ps).

    The names are only used for output (console printing and plotting), and are optional.
    The conversion factors and kB are, on the other hand, used in computations and need
    to be given.
    """
    def __init__(self, kb, energy_conversion, length_conversion, volume_conversion,
                 temperature_conversion, pressure_conversion, time_conversion,
                 energy_str='ENE', length_str='LEN', volume_str='VOL',
                 temperature_str='TEMP', pressure_str='PRESS', time_str='TIME'):

        self.__kb = float(kb)
        self.__energy_str = str(energy_str)
        self.__energy_conversion = float(energy_conversion)
        self.__length_str = str(length_str)
        self.__length_conversion = float(length_conversion)
        self.__volume_str = str(volume_str)
        self.__volume_conversion = float(volume_conversion)
        self.__temperature_str = str(temperature_str)
        self.__temperature_conversion = float(temperature_conversion)
        self.__pressure_str = str(pressure_str)
        self.__pressure_conversion = float(pressure_conversion)
        self.__time_str = str(time_str)
        self.__time_conversion = float(time_conversion)

    @staticmethod
    def __parsers():
        from . import GromacsParser
        return {
            'GROMACS': GromacsParser
        }

    @classmethod
    def units(cls, name=None):
        if name is None:
            return cls.__parsers().keys()

        if name in cls.__parsers():
            return cls.__parsers()[name].units()
        else:
            raise KeyError('Name ' + name + ' does not match a registred unit type.')

    def __eq__(self, other):
        if not isinstance(other, UnitData):
            return False

        return (self.kb == other.kb and
                self.energy_conversion == other.energy_conversion and
                self.length_conversion == other.length_conversion and
                self.volume_conversion == other.volume_conversion and
                self.temperature_conversion == other.temperature_conversion and
                self.pressure_conversion == other.pressure_conversion and
                self.time_conversion == other.time_conversion)

    @property
    def kb(self):
        """float: The value of the Boltzmann constant"""
        return self.__kb

    @property
    def energy_str(self):
        """str: Energy unit"""
        return self.__energy_str

    @property
    def length_str(self):
        """str: Length unit"""
        return self.__length_str

    @property
    def volume_str(self):
        """str: Volume unit"""
        return self.__volume_str

    @property
    def temperature_str(self):
        """str: Temperature unit"""
        return self.__temperature_str

    @property
    def pressure_str(self):
        """str: Pressure unit"""
        return self.__pressure_str

    @property
    def time_str(self):
        """str: Time unit"""
        return self.__time_str

    @property
    def energy_conversion(self):
        """float: Energy conversion factor, 1 energy_unit = energy_conversion * kJ/mol"""
        return self.__energy_conversion

    @property
    def length_conversion(self):
        """float: Length conversion factor, 1 length_unit = length_conversion * nm"""
        return self.__length_conversion

    @property
    def volume_conversion(self):
        """float: Volume conversion factor, 1 volume_unit = volume_conversion * nm^3"""
        return self.__volume_conversion

    @property
    def temperature_conversion(self):
        """float: Temperature conversion factor, 1 temperature_unit = temperature_conversion * K"""
        return self.__temperature_conversion

    @property
    def pressure_conversion(self):
        """float: Pressure conversion factor, 1 pressure_unit = pressure_conversion * bar"""
        return self.__pressure_conversion

    @property
    def time_conversion(self):
        """float: Time conversion factor, 1 time_unit = time_conversion * ps"""
        return self.__time_conversion



#
# This file is part of the GROMACS molecular simulation package.
#
# Copyright 2020- The GROMACS Authors
# and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
# Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
#
# GROMACS is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# GROMACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with GROMACS; if not, see
# https://www.gnu.org/licenses, or write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
#
# If you want to redistribute modifications to GROMACS, please
# consider that scientific software is very special. Version
# control is crucial - bugs must be traceable. We will be happy to
# consider code for inclusion in the official distribution, but
# derived work must not be called official GROMACS. Details are found
# in the README & COPYING files - if they are missing, get the
# official version at https://www.gromacs.org.
#
# To help us fund GROMACS development, we humbly ask that you cite
# the research papers on the package. Check out https://www.gromacs.org.

"""A utility module to help manage the matrix of configurations for CI testing and build containers.

When called as a stand alone script, prints a Docker image name based on the
command line arguments. The Docker image name is of the form used in the GROMACS
CI pipeline jobs.

Example::

    $ python3 -m utility --llvm --doxygen
    gromacs/ci-ubuntu-24.04-llvm-14-docs

See Also:
    :file:`buildall.sh`

As a module, provides importable argument parser and docker image name generator.

Note that the parser is created with ``add_help=False`` to make it friendly as a
parent parser, but this means that you must derive a new parser from it if you
want to see the full generated command line help.

Example::

    import utility.parser
    # utility.parser does not support `-h` or `--help`
    parser = argparse.ArgumentParser(
        description='GROMACS CI image creation script',
        parents=[utility.parser])
    # ArgumentParser(add_help=True) is default, so parser supports `-h` and `--help`

See Also:
    :file:`scripted_gmx_docker_builds.py`

Authors:
    * Paul Bauer <paul.bauer.q@gmail.com>
    * Eric Irrgang <ericirrgang@gmail.com>
    * Joe Jordan <e.jjordan12@gmail.com>
    * Mark Abraham <mark.j.abraham@gmail.com>
    * Gaurav Garg <gaugarg@nvidia.com>

"""

import argparse


parser = argparse.ArgumentParser(
    description="GROMACS CI image slug options.", add_help=False
)
"""A parent parser for tools referencing image parameters.

This argparse parser is defined for convenience and may be used to partially initialize
parsers for tools.

.. warning:: Do not modify this parser.

    Instead, inherit from it with the *parents* argument to :py:class:`argparse.ArgumentParser`
"""

parser.add_argument(
    "--cmake",
    nargs="*",
    type=str,
    default=["3.28.0", "3.30.3", "4.0.5"],
    help="Selection of CMake version to provide to base image. (default: %(default)s)",
)

compiler_group = parser.add_mutually_exclusive_group()
compiler_group.add_argument(
    "--gcc",
    type=int,
    default=11,
    help="Select GNU compiler tool chain. (default: %(default)s) "
    "Some checking is implemented to avoid incompatible combinations",
)
compiler_group.add_argument(
    "--llvm",
    type=str,
    nargs="?",
    const="14",
    default=None,
    help="Select LLVM compiler tool chain. "
    "Some checking is implemented to avoid incompatible combinations",
)
# Note that oneAPI packages don't bump their version numbers every
# version of the umbrella release, so we may need to specify such
# package versions from time to time.
compiler_group.add_argument(
    "--oneapi",
    type=str,
    nargs="?",
    const="2024.0.0",
    default=None,
    help="Select Intel oneAPI package version.",
)
compiler_group.add_argument(
    "--intel-llvm",
    type=str,
    nargs="?",
    const="2022-06",
    default=None,
    help="Select Intel LLVM release (GitHub tag).",
)

linux_group = parser.add_mutually_exclusive_group()
linux_group.add_argument(
    "--ubuntu",
    type=str,
    nargs="?",
    const="24.04",
    default="24.04",
    help="Select Ubuntu Linux base image. (default: %(default)s)",
)
linux_group.add_argument(
    "--centos",
    type=str,
    nargs="?",
    const="7",
    default=None,
    help="Select Centos Linux base image.",
)

parser.add_argument(
    "--cuda",
    type=str,
    nargs="?",
    const="12.1",
    default=None,
    help="Select a CUDA version for a base Linux image from NVIDIA.",
)

parser.add_argument(
    "--mpi",
    type=str,
    nargs="?",
    const="openmpi",
    default=None,
    help="Enable MPI (default disabled) and optionally select distribution (default: openmpi)",
)

parser.add_argument(
    "--tsan",
    type=str,
    nargs="?",
    const="llvm",
    default=None,
    help="Build special compiler versions with TSAN OpenMP support",
)

parser.add_argument(
    "--adaptivecpp",
    type=str,
    nargs="?",
    default=None,
    help="Select AdaptiveCpp repository tag/commit/branch.",
)

parser.add_argument(
    "--rocm",
    type=str,
    nargs="?",
    const="3.5.1",
    default=None,
    help="Select AMD compute engine version.",
)

parser.add_argument(
    "--intel-compute-runtime",
    action="store_true",
    default=False,
    help="Include Intel Compute Runtime.",
)

parser.add_argument(
    "--oneapi-plugin-nvidia",
    action="store_true",
    default=False,
    help="Install Codeplay oneAPI NVIDIA plugin.",
)

parser.add_argument(
    "--oneapi-plugin-amd",
    action="store_true",
    default=False,
    help="Install Codeplay oneAPI AMD plugin.",
)

parser.add_argument(
    "--clfft",
    type=str,
    nargs="?",
    const="master",
    default=None,
    help="Add external clFFT libraries to the build image",
)

parser.add_argument(
    "--heffte",
    type=str,
    nargs="?",
    default=None,
    help="Select heffte repository tag/commit/branch.",
)

parser.add_argument(
    "--nvhpcsdk",
    type=str,
    nargs="?",
    default=None,
    help="Select NVIDIA HPC SDK version.",
)

parser.add_argument(
    "--doxygen",
    type=str,
    nargs="?",
    const="1.8.5",
    default=None,
    help="Add doxygen environment for documentation builds. Also adds other requirements needed for final docs images.",
)

parser.add_argument(
    "--cp2k",
    type=str,
    nargs="?",
    const="9.1",
    default=None,
    help="Add build environment for CP2K QM/MM support",
)

parser.add_argument(
    "--hdf5",
    action="store_true",
    default=False,
    help="Install an HDF5 library.",
)

parser.add_argument(
    "--plumed",
    action="store_true",
    default=False,
    help="Install PLUMED library.",
)

parser.add_argument(
    "--cross",
    type=str,
    nargs="?",
    const="aarch64",
    default=None,
    choices=["aarch64", "riscv64"],
    help="Add cross-compilation support for the specified architecture with QEMU emulation",
)

parser.add_argument(
    "--libtorch",
    type=str,
    nargs="?",
    const="2.9.0",
    default=None,
    help="Add build environment for LibTorch support",
)

# Supported Python versions for maintained branches.
_python_versions = ["3.9.13", "3.12.5"]
parser.add_argument(
    "--venvs",
    nargs="*",
    type=str,
    default=_python_versions,
    help='List of Python versions ("major.minor.patch") for which to install venvs. (default: %(default)s)',
)


def image_name(configuration: argparse.Namespace) -> str:
    """Generate docker image name.

    Image names have the form ``ci-<slug>``, where the configuration slug has the form::

        <distro>-<version>-<compiler>-<major version>[-<gpusdk>-<version>][-<use case>]

    This function also applies an appropriate Docker image repository prefix.

    Arguments:
        configuration: Docker image configuration as described by the parsed arguments.

    """
    elements = []
    for distro in ("centos", "ubuntu"):
        version = getattr(configuration, distro, None)
        if version is not None:
            elements.append(distro + "-" + version)
            break
    for compiler in ("llvm", "intel_llvm", "oneapi", "gcc"):
        version = getattr(configuration, compiler, None)
        if version is not None:
            version = (
                str(version).split(".")[0] if compiler != "oneapi" else str(version)
            )
            elements.append(compiler + "-" + version)
            break
    for gpusdk in ("cuda", "adaptivecpp"):
        version = getattr(configuration, gpusdk, None)
        if version is not None:
            elements.append(gpusdk + "-" + version)
    if configuration.intel_compute_runtime:
        elements.append("intel-compute-runtime")
    if configuration.rocm is not None:
        elements.append("rocm-" + configuration.rocm)
    if configuration.cp2k is not None:
        elements.append("cp2k-" + configuration.cp2k)

    # Check for special cases
    # The following attribute keys indicate the image is built for the named
    # special use case.
    cases = {"doxygen": "docs", "tsan": "tsan"}
    for attr in cases:
        value = getattr(configuration, attr, None)
        if value is not None:
            elements.append(cases[attr])

    # Add cross-compilation target if specified
    if configuration.cross is not None:
        elements.append(f"cross-{configuration.cross}")
    slug = "-".join(elements)
    # we are using the GitLab container registry to store the images
    # to get around issues with pulling them repeatedly from DockerHub
    # and running into the image pull limitation there.
    return "registry.gitlab.com/gromacs/gromacs/ci-" + slug


if __name__ == "__main__":
    args = argparse.ArgumentParser(parents=[parser]).parse_args()
    print(image_name(args))



#
# This file is part of the GROMACS molecular simulation package.
#
# Copyright 2019- The GROMACS Authors
# and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
# Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
#
# GROMACS is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# GROMACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with GROMACS; if not, see
# https://www.gnu.org/licenses, or write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
#
# If you want to redistribute modifications to GROMACS, please
# consider that scientific software is very special. Version
# control is crucial - bugs must be traceable. We will be happy to
# consider code for inclusion in the official distribution, but
# derived work must not be called official GROMACS. Details are found
# in the README & COPYING files - if they are missing, get the
# official version at https://www.gromacs.org.
#
# To help us fund GROMACS development, we humbly ask that you cite
# the research papers on the package. Check out https://www.gromacs.org.

"""
Exceptions and Warnings raised by gmxapi module operations.

Errors, warnings, and other exceptions used in the GROMACS
Python package are defined in the `exceptions` submodule.

The gmxapi Python package defines a root exception,
exceptions.Error, from which all Exceptions thrown from
within the module should derive. If a published component of
the gmxapi package throws an exception that cannot be caught
as a gmxapi.exceptions.Error, please report the bug.
"""

__all__ = [
    "ApiError",
    "DataShapeError",
    "Error",
    "FeatureNotAvailableError",
    "MissingImplementationError",
    "ProtocolError",
    "TypeError",
    "UsageError",
    "ValueError",
    "Warning",
]


class Error(Exception):
    """Base exception for gmx.exceptions classes."""


class Warning(Warning):
    """Base warning class for gmx.exceptions."""


class ApiError(Error):
    """An API operation was attempted with an incompatible object."""


class DataShapeError(Error):
    """An object has an incompatible shape.

    This exception does not imply that the Type or any other aspect of the data
    has been checked.
    """


class MissingImplementationError(Error):
    """Specified feature is not implemented in the current code.

    This exception indicates that the implemented code does not support the
    API as specified. The calling code has used valid syntax, as documented for
    the API, but has reached incompletely implemented code, which should be
    considered a bug.

    .. versionchanged:: 0.3
        Named changed to avoid conflict with built-in :py:class:`NotImplementedError` exception
    """

    # May be useful for error checking in base classes or as a development tool
    # to avoid releasing incomplete implementations (e.g. overlooked "To do"s)


class FeatureNotAvailableError(Error):
    """Requested feature not available in the current environment.

    This exception will usually indicate an issue with the user's environment or
    run time details. There may be a missing optional dependency, which should
    be specified in the exception message.
    """


class ProtocolError(Error):
    """Unexpected API behavior or protocol violation.

    This exception generally indicates a gmxapi bug, since it should only
    occur through incorrect assumptions or misuse of API implementation internals.
    """


class TypeError(Error):
    """Incompatible type for gmxapi data.

    Reference datamodel.rst for more on gmxapi data typing.
    """


class UsageError(Error):
    """Unsupported syntax or call signatures.

    Generic usage error for gmxapi module.
    """


class ValueError(Error):
    """A user-provided value cannot be interpreted or doesn't make sense."""
